# coding: utf-8

"""
    Thoth User API

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: 0.7.0-dev

    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""

import pprint
import re  # noqa: F401

import six

class AnalysisResultResponseResultDeb(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """
    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'arch': 'str',
        'name': 'str',
        'version': 'str'
    }

    attribute_map = {
        'arch': 'arch',
        'name': 'name',
        'version': 'version'
    }

    def __init__(self, arch=None, name=None, version=None):  # noqa: E501
        """AnalysisResultResponseResultDeb - a model defined in Swagger"""  # noqa: E501
        self._arch = None
        self._name = None
        self._version = None
        self.discriminator = None
        self.arch = arch
        self.name = name
        self.version = version

    @property
    def arch(self):
        """Gets the arch of this AnalysisResultResponseResultDeb.  # noqa: E501

        Package architecture  # noqa: E501

        :return: The arch of this AnalysisResultResponseResultDeb.  # noqa: E501
        :rtype: str
        """
        return self._arch

    @arch.setter
    def arch(self, arch):
        """Sets the arch of this AnalysisResultResponseResultDeb.

        Package architecture  # noqa: E501

        :param arch: The arch of this AnalysisResultResponseResultDeb.  # noqa: E501
        :type: str
        """
        if arch is None:
            raise ValueError("Invalid value for `arch`, must not be `None`")  # noqa: E501

        self._arch = arch

    @property
    def name(self):
        """Gets the name of this AnalysisResultResponseResultDeb.  # noqa: E501

        Name of the Debian package  # noqa: E501

        :return: The name of this AnalysisResultResponseResultDeb.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this AnalysisResultResponseResultDeb.

        Name of the Debian package  # noqa: E501

        :param name: The name of this AnalysisResultResponseResultDeb.  # noqa: E501
        :type: str
        """
        if name is None:
            raise ValueError("Invalid value for `name`, must not be `None`")  # noqa: E501

        self._name = name

    @property
    def version(self):
        """Gets the version of this AnalysisResultResponseResultDeb.  # noqa: E501

        Version identifier of the Debian package  # noqa: E501

        :return: The version of this AnalysisResultResponseResultDeb.  # noqa: E501
        :rtype: str
        """
        return self._version

    @version.setter
    def version(self, version):
        """Sets the version of this AnalysisResultResponseResultDeb.

        Version identifier of the Debian package  # noqa: E501

        :param version: The version of this AnalysisResultResponseResultDeb.  # noqa: E501
        :type: str
        """
        if version is None:
            raise ValueError("Invalid value for `version`, must not be `None`")  # noqa: E501

        self._version = version

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(AnalysisResultResponseResultDeb, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, AnalysisResultResponseResultDeb):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        return not self == other



#!/usr/bin/env python3
# thamos
# Copyright(C) 2018 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Implementation discovery methods to detect the current environment and its configuration."""

import os
import sys
import logging
import typing
import sysconfig

from typing import Dict, Union

from thoth.common import map_os_name
from thoth.common import normalize_os_version

import distro
import click
from thoth.analyzer import run_command

_LOGGER = logging.getLogger(__name__)
_PROC_CPU_INFO = "/proc/cpuinfo"


def discover_cuda_version(interactive: bool = False) -> typing.Optional[str]:
    """Check for CUDA version, if no CUDA is installed, return None."""
    if bool(int(os.getenv("THAMOS_DISABLE_CUDA", 0))):
        _LOGGER.debug(
            "Disabling CUDA based on THAMOS_DISABLE_CUDA environment variable that is set to %r",
            os.environ["THAMOS_DISABLE_CUDA"],
        )
        return None

    result = run_command("nvcc --version", raise_on_error=False)
    if result.return_code != 0:
        _LOGGER.debug("No CUDA version detected")
        _LOGGER.debug(
            "Unable to detect CUDA version - nvcc returned non-zero version: %s",
            result.to_dict(),
        )
        return None

    lines = result.stdout.splitlines()
    version_info = lines[-1].split(",")
    if len(version_info) != 3:
        _LOGGER.debug(
            "Unable to detect CUDA version from nvcc output: %r", result.stdout
        )
        return None

    cuda_version = version_info[1].strip()[len("release ") :]

    if interactive:
        cuda_version = click.prompt("Please select CUDA version", default=cuda_version)

    _LOGGER.debug("Detected CUDA version: %r", cuda_version)
    return cuda_version


def discover_distribution() -> tuple:
    """Get distribution identifier and distribution version."""
    distribution, version, *_ = distro.linux_distribution(full_distribution_name=False)
    _LOGGER.debug("Detected running %r in version %r", distribution, version)
    os_name = map_os_name(distribution)
    return os_name, normalize_os_version(os_name, version)


def discover_python_version() -> str:
    """Discover Python version in which we run in."""
    python_version = f"{sys.version_info.major}.{sys.version_info.minor}"
    _LOGGER.debug("Detected running Python %r", python_version)
    return python_version


def discover_cpu() -> Dict[str, Union[str, int, None]]:
    """Discover CPU model, model name and family."""
    result = {
        "cpu_family": None,
        "cpu_model": None,
        "cpu_model_name": None,
    }  # type: Dict[str, Union[str, int, None]]

    try:
        with open(_PROC_CPU_INFO, "r") as cpu_info_file:
            content = cpu_info_file.read().splitlines()

        for line in content:
            if line.startswith("model name") and result["cpu_model_name"] is None:
                try:
                    result["cpu_model_name"] = line.split(":")[1].strip()
                except Exception as exc:
                    _LOGGER.warning(
                        "Failed to obtain CPU model name from %s: %s",
                        str(exc),
                        _PROC_CPU_INFO,
                    )
            elif line.startswith("model\t") and result["cpu_model"] is None:
                try:
                    result["cpu_model"] = int(line.split(":")[1])
                except Exception as exc:
                    _LOGGER.warning(
                        "Failed to obtain CPU model from %s: %s",
                        str(exc),
                        _PROC_CPU_INFO,
                    )
            elif line.startswith("cpu family") and result["cpu_family"] is None:
                try:
                    result["cpu_family"] = int(line.split(":")[1])
                except Exception as exc:
                    _LOGGER.warning(
                        "Failed to obtain CPU family from %s: %s",
                        str(exc),
                        _PROC_CPU_INFO,
                    )

    except Exception as exc:
        _LOGGER.exception("Failed to obtain CPU specific information: %s", str(exc))

    if result["cpu_model_name"] is None:
        # Assign a text representation - unknown for config file.
        result["cpu_model_name"] = "Unknown"

    _LOGGER.debug(
        "Detected CPU: %s", ", ".join((f"{k}: {v}" for k, v in result.items()))
    )
    return result


def discover_platform() -> str:
    """Discover platform used."""
    platform = sysconfig.get_platform()
    _LOGGER.debug("Detected running platform %r", platform)
    return platform


def discover_base_image() -> typing.Optional[str]:
    """Discover base image and its version."""
    # IMAGE_NAME and IMAGE_TAG injected by AICoE-CI take precedence over Thoth s2i.
    base_image_name = os.getenv("IMAGE_NAME", os.getenv("THOTH_S2I_NAME"))
    base_image_version = os.getenv("IMAGE_TAG")
    if base_image_version is None:
        base_image_version = os.getenv("THOTH_S2I_VERSION")
        # Add `v' to the version information for Thoth specific environment variable.
        base_image_version = f"v{base_image_version}" if base_image_version else None

    if base_image_name and base_image_version:
        base_image = f"{base_image_name}:{base_image_version}"
        _LOGGER.debug("Detected base image %r", base_image)
        return base_image
    elif base_image_name:
        _LOGGER.warning(
            "Discovered running inside %r but no base image version provided",
            base_image_name,
        )
    elif base_image_version:
        _LOGGER.warning(
            "Discovered running inside a base image in version %r but no base image name provided",
            base_image_version,
        )

    return None


def discover_all() -> typing.Dict[str, typing.Any]:
    """Discover all the entries used in Thoth's configuration file."""
    # XXX: missing autodiscovery for OpenBLAS, OpenMPI, cuDNN, MKL, hardware.gpu_model
    result = {
        "cuda_version": discover_cuda_version(),
        "python_version": discover_python_version(),
        "platform": discover_platform(),
        "base_image": discover_base_image(),
        "hardware": discover_cpu(),
    }

    os_name, os_version = discover_distribution()
    result["operating_system"] = {
        "name": os_name,
        "version": os_version,
    }

    return result



#!/usr/bin/env python3
# thamos
# Copyright(C) 2022 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Configuration for Thamos CLI using rich-click."""

import os
import sys
import rich_click as click


def init_rich_click() -> None:
    """Initialize rich-click."""
    click.rich_click.ERRORS_EPILOGUE = (
        "Check Thamos documentation: https://thoth-station.ninja/docs/developers/thamos"
    )
    click.rich_click.HEADER_TEXT = "Thoth homepage: https://thoth-station.ninja"
    click.rich_click.FOOTER_TEXT = (
        "Check Thamos documentation: https://thoth-station.ninja/docs/developers/thamos"
    )
    click.rich_click.USE_RICH_MARKUP = True

    tool = os.path.basename(sys.argv[0])

    click.rich_click.COMMAND_GROUPS = {
        tool: [
            {
                "name": "Thoth resolver",
                "commands": [
                    "advise",
                    "environments",
                    "graph",
                    "images",
                    "indexes",
                    "log",
                    "provenance-check",
                    "status",
                    "whatprovides",
                ],
            },
            {
                "name": "Application & Requirements",
                "commands": [
                    "add",
                    "discover",
                    "install",
                    "purge",
                    "remove",
                    "run",
                    "venv",
                    "verify",
                ],
            },
            {
                "name": "Runtime environments",
                "commands": [
                    "check",
                    "config",
                    "list",
                    "show",
                ],
            },
            {
                "name": "Information",
                "commands": [
                    "support",
                    "version",
                ],
            },
        ],
    }



#!/usr/bin/env python3
# thamos
# Copyright(C) 2018 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""CLI and library configuration."""

import logging
import os
import re
import sys
from typing import Any
from typing import Dict
from typing import Optional
from typing import List
from urllib.parse import urljoin
from jsonschema import validate

import click
import requests
import yaml

from thoth.common import map_os_name
from thoth.common import normalize_os_version
from thoth.common import RuntimeEnvironment
from thoth.python import Pipfile
from thoth.python import Project
from virtualenv import cli_run as virtualenv_cli_run

from .utils import workdir
from .discover import discover_cpu
from .discover import discover_cuda_version
from .discover import discover_distribution
from .discover import discover_python_version
from .discover import discover_platform
from .discover import discover_base_image
from .exceptions import NoApiSupported
from .exceptions import NoRuntimeEnvironmentError
from .exceptions import RuntimeEnvironmentExistsError
from .exceptions import ConfigurationError
from .exceptions import NoProjectDirError
from .exceptions import ServiceUnavailable

_LOGGER = logging.getLogger(__name__)
_THAMOS_DISABLE_TLS_WARNING = bool(int(os.getenv("THAMOS_DISABLE_TLS_WARNING", 0)))
_API_CONNECTION_TIMEOUT = int(os.getenv("THAMOS_API_CONNECTION_TIMEOUT", 5))

# The schema is enforcing all the options. This will make sure the right version of Thamos is
# installed and no configuration options are silently ignored.
_CONFIG_RUNTIME_ENVIRONMENT_SCHEMA = {
    "type": "object",
    "properties": {
        "cuda_version": {"type": ["string", "null"]},
        "openblas_version": {"type": ["string", "null"]},
        "openmpi_version": {"type": ["string", "null"]},
        "cudnn_version": {"type": ["string", "null"]},
        "mkl_version": {"type": ["string", "null"]},
        "base_image": {"type": ["string", "null"]},
        "labels": {"type": ["object", "null"]},
        "hardware": {
            "type": "object",
            "properties": {
                "cpu_family": {"type": "integer"},
                "cpu_model": {"type": "integer"},
                "gpu_model": {"type": ["string", "null"]},
            },
            "required": [],
            "additionalProperties": False,
        },
        "name": {"type": "string", "pattern": r"^[a-zA-Z0-9_-]+$"},
        "operating_system": {
            "type": "object",
            "properties": {
                "name": {"type": ["string", "null"]},
                "version": {"type": ["string", "null"]},
            },
            "required": ["name", "version"],
            "additionalProperties": False,
        },
        "platform": {"type": ["string", "null"]},
        "python_version": {
            "type": "string",
            "pattern": r"^[0-9]+\.[0-9]+$",
        },
        "recommendation_type": {
            "type": "string",
            "enum": [
                "latest",
                "stable",
                "performance",
                "security",
                "testing",
            ],
        },
    },
    "additionalProperties": False,
    "required": ["name"],
}
_CONFIG_SCHEMA = {
    "type": "object",
    "properties": {
        "host": {"type": "string", "format": "hostname"},
        "virtualenv": {"type": "boolean"},
        "tls_verify": {"type": "boolean"},
        "requirements_format": {
            "type": "string",
            "enum": ["pipenv", "pip", "pip-tools"],
        },
        "runtime_environments": {
            "type": "array",
            "items": _CONFIG_RUNTIME_ENVIRONMENT_SCHEMA,
            "minItems": 1,
        },
        "managers": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "name": {"type": "string"},
                },
                "additionalProperties": True,
                "required": ["name"],
            },
        },
    },
    "required": ["host", "runtime_environments"],
}


class _Configuration:
    """Handling of Thoth's configuration."""

    DATA_DIR = os.path.join(os.path.dirname(os.path.realpath(__file__)), "data")
    DEFAULT_THOTH_CONFIG = os.path.join(DATA_DIR, "defaultThoth.yaml")
    CONFIG_NAME = ".thoth.yaml"
    REQUIREMENTS_FORMATS = frozenset(("pip", "pip-tools", "pip-compile", "pipenv"))
    _DEFAULT_REQUIREMENTS_FORMAT = "pipenv"
    _TLS_WARNING_LOGGED = False

    def __init__(self):
        """Construct configuration instance."""
        self._configuration = None
        self._index_aicoe = None
        # Thoth instance to be used when explicitly said by user - the one stated in
        # configuration file will be omitted.
        self.explicit_host = None
        self.tls_verify = None
        self._api_url = None
        self._thoth_search_ui_url = None

    @property
    def api_url(self):
        """Get URL to Thoth's API."""
        if not self._api_url:
            self._api_url = self.api_discovery(self.content["host"])

        return self._api_url

    @property
    def thoth_search_ui_url(self):
        if not self._thoth_search_ui_url:
            self.api_discovery(self.content["host"])

        return self._thoth_search_ui_url

    @property
    def content(self):
        """Access configuration."""
        if not self._configuration:
            self.load_config()

        return self._configuration

    @property
    def requirements_format(self) -> str:
        """Check requirements_format in configuration."""
        requirements_format = self.content.get("requirements_format") or "pipenv"

        if not isinstance(requirements_format, str):
            raise ConfigurationError(
                "The data type for requirements_format should be str"
            )

        if requirements_format not in ("pip", "pip-tools", "pipenv"):
            raise ValueError(
                f"Unknown configuration option for requirements format: {requirements_format!r}"
            )

        return requirements_format

    def get_thoth_version(self) -> str:
        """Get version of Thoth backend."""
        _LOGGER.debug(
            "Contacting Thoth at %r to receive version information", self.api_url
        )
        response = requests.head(
            self.api_url, verify=self.tls_verify, timeout=_API_CONNECTION_TIMEOUT
        )
        response.raise_for_status()
        return response.headers.get("X-Thoth-Version", "Not Available")

    def get_virtualenv_path(
        self, runtime_environment: Optional[str] = None
    ) -> Optional[str]:
        """Get path to a virtual environment."""
        if not self.content.get("virtualenv", False):
            return None

        if sys.base_prefix != sys.prefix:
            _LOGGER.warning(
                "Detected running in a virtual environment, no managed virtual environment will be used"
            )
            return None

        return os.path.join(self.get_overlays_directory(runtime_environment), ".venv")

    def create_virtualenv(self, runtime_environment: Optional[str] = None) -> None:
        """Create a virtual environment for the given runtime environment."""
        virtualenv_path = self.get_virtualenv_path(runtime_environment)
        if virtualenv_path is None:
            raise ConfigurationError("No virtual environment configured")

        virtualenv_args = [virtualenv_path]
        runtime_environment_dict = self.get_runtime_environment(runtime_environment)
        python_version = runtime_environment_dict.get("python_version")
        if python_version:
            virtualenv_args.extend(["--python", python_version])

        _LOGGER.info("Creating virtual environment")
        _LOGGER.debug(
            "Virtual environment will be created in %r using %r",
            virtualenv_path,
            virtualenv_args,
        )
        virtualenv_cli_run(virtualenv_args)  # Raises on any error.

    def config_file_exists(self) -> bool:
        """Check if configuration file exists."""
        try:
            with workdir(self.CONFIG_NAME):
                return True
        except NoProjectDirError:
            return False

    def load_config_from_string(self, config_str: str) -> None:
        """Load configuration from a string."""
        if int(os.getenv("THAMOS_CONFIG_EXPAND_ENV", 0)):
            _LOGGER.info("Expanding configuration file based on environment variables")
            config_str = config_str.format(**os.environ)

        self._configuration = yaml.safe_load(config_str)

    def load_config_from_file(self, config_path: str) -> None:
        """Load configuration from a file."""
        with open(config_path, "r") as config_file:
            self._configuration = config_file.read()

            if int(os.getenv("THAMOS_CONFIG_EXPAND_ENV", 0)):
                _LOGGER.info(
                    "Expanding configuration file based on environment variables"
                )
                self._configuration = self._configuration.format(**os.environ)

            self._configuration = yaml.safe_load(self._configuration)
            self.tls_verify = (
                self._configuration.get("tls_verify", True)
                if self.tls_verify is None
                else self.tls_verify
            )

    def reset_config(self) -> None:
        """Discard loaded config in memory."""
        self._configuration = None

    def load_config(self, force: bool = False) -> None:
        """Load configuration from a file."""
        if not self._configuration and not force:
            with workdir(config.CONFIG_NAME):
                self.load_config_from_file(config.CONFIG_NAME)

    def save_config(self, path: Optional[str] = None) -> None:
        """Save the configuration to disc."""
        if path:
            with open(path, "w") as f:
                yaml.dump(self.content, f)
            _LOGGER.debug("Configuration changes written to %r", path)
        else:
            with workdir(config.CONFIG_NAME), open(config.CONFIG_NAME, "w") as f:
                yaml.dump(self.content, f)
                _LOGGER.debug(
                    "Configuration changes written to %r",
                    os.path.join(os.getcwd(), config.CONFIG_NAME),
                )

    def create_default_config(
        self, template: str = None, nowrite: bool = False
    ) -> Optional[dict]:
        """Place default configuration into the current directory."""
        if not os.path.isdir(".git"):
            _LOGGER.warning("Configuration file is not created in the root of git repo")

        template = template or self.DEFAULT_THOTH_CONFIG
        _LOGGER.debug("Reading configuration from %r", template)
        with open(template, "r") as default_config_file:
            default_config = default_config_file.read()

        _LOGGER.info("Discovering host runtime environment")

        cpu_info = discover_cpu()
        cuda_version = discover_cuda_version()
        # Add quotes for textual representation in the config file.
        cuda_version = f"'{cuda_version}'" if cuda_version is not None else "null"
        os_name, os_version = discover_distribution()
        os_name = map_os_name(os_name)
        os_version = normalize_os_version(os_name, os_version)
        python_version = discover_python_version()
        platform = discover_platform()
        base_image = discover_base_image()
        base_image = base_image if base_image is not None else "null"

        runtime_environment_name = re.sub(
            r"[^0-9a-zA-Z-]", "-", f"{os_name}-{os_version}"
        )

        requirements_format = os.getenv(
            "THAMOS_REQUIREMENTS_FORMAT", self._DEFAULT_REQUIREMENTS_FORMAT
        )
        if requirements_format not in self.REQUIREMENTS_FORMATS:
            # This avoids possibly dangerous environment variable expansion.
            _LOGGER.warning(
                "Unknown requirements format specified, forcing %r: %r",
                self._DEFAULT_REQUIREMENTS_FORMAT,
                requirements_format,
            )
            requirements_format = self._DEFAULT_REQUIREMENTS_FORMAT

        expand_env = bool(int(os.getenv("THAMOS_CONFIG_EXPAND_ENV", 0)))
        default_config = default_config.format(
            runtime_environment_name=runtime_environment_name,
            cuda_version=cuda_version,
            os_name=os_name,
            os_version=os_version,
            python_version=python_version,
            platform=platform,
            requirements_format=requirements_format,
            base_image=base_image,
            cpu_family=cpu_info["cpu_family"] or "null",
            cpu_model=cpu_info["cpu_model"] or "null",
            cpu_model_name=cpu_info["cpu_model_name"] or "CPU detection failed",
            **(dict(os.environ) if expand_env else {}),
        )

        if not nowrite:
            _LOGGER.debug(
                "Writing configuration file to %r",
                os.path.join(os.getcwd(), self.CONFIG_NAME),
            )

            with open(self.CONFIG_NAME, "w") as config_file:
                config_file.write(default_config)
            return None
        else:
            return yaml.safe_load(default_config)

    @staticmethod
    def open_config_file():
        """Open Thoth's configuration file."""
        with workdir(config.CONFIG_NAME):
            _LOGGER.debug("Opening configuration file %r", config.CONFIG_NAME)
            click.edit(filename=config.CONFIG_NAME)

    def list_runtime_environments(self):
        """List available runtime environments."""
        return self.content.get("runtime_environments", [])

    def set_runtime_environment(
        self, runtime_environment: Dict[str, Any], force: bool = False
    ) -> None:
        """Add a runtime environment entry, overrides already existing one if force was set."""
        try:
            validate(runtime_environment, _CONFIG_RUNTIME_ENVIRONMENT_SCHEMA)
        except Exception:
            _LOGGER.error("Failed to validate runtime environment entry against schema")
            raise

        try:
            existing = self.get_runtime_environment(runtime_environment["name"])
        except NoRuntimeEnvironmentError:
            self.content.setdefault("runtime_environments", []).append(
                runtime_environment
            )
        else:
            if force:
                existing.clear()
                existing.update(runtime_environment)
            else:
                raise RuntimeEnvironmentExistsError(
                    f"Runtime environment {runtime_environment['name']!r} already exists"
                )

    def get_runtime_environment(self, name: Optional[str] = None) -> Dict[str, Any]:
        """Get runtime environment, retrieve the first runtime environment (the default one) if no name is provided."""
        content = self.content
        if "runtime_environments" not in content:
            raise NoRuntimeEnvironmentError(
                "No runtime environment configuration stated in the configuration file "
                "under 'runtime_environments' configuration entry"
            )

        if not isinstance(content["runtime_environments"], list):
            raise ConfigurationError(
                "The data type for requirements_format should be list"
            )

        to_return = None
        seen_names = set()
        for idx, runtime_environment in enumerate(content["runtime_environments"]):
            if not isinstance(runtime_environment, dict):
                raise ConfigurationError(
                    "Unknown runtime configuration entry, runtime configuration should be "
                    "a dictionary; got: %r",
                    runtime_environment,
                )

            # We explicitly iterate over all entries to perform the following sanity checks.
            current_name = runtime_environment.get("name")
            if current_name is not None and current_name in seen_names:
                raise ConfigurationError(
                    "Multiple configuration options with name %r found in the configuration file",
                    current_name,
                )

            if idx > 0 and current_name is None:
                raise ConfigurationError(
                    "Assign explicitly name for each configuration entry if there are multiple "
                    "runtime configuration options to distinguish between them"
                )

            if current_name is not None:
                seen_names.add(current_name)

            if name is None and idx == 0:
                # Return the first one by default.
                to_return = runtime_environment
            elif current_name == name:
                # Return by name.
                to_return = runtime_environment

        if to_return is None:
            if name is not None:
                raise NoRuntimeEnvironmentError(
                    f"No runtime environment with name {name!r} was found in the configuration file; "
                    f"configured runtime environment names: {','.join(seen_names)}"
                )

            raise NoRuntimeEnvironmentError(
                "No runtime environment configuration was found"
            )

        return to_return

    def api_discovery(self, host: str) -> str:
        """Discover API versions available, return the most recent one supported by client and server."""
        api_url = urljoin("https://" + host, "api/v1")
        self.tls_verify = (
            self.tls_verify
            if self.tls_verify is not None
            else self.content.get("tls_verify", True)
        )

        if (
            not self.tls_verify
            and not _THAMOS_DISABLE_TLS_WARNING
            and not self._TLS_WARNING_LOGGED
        ):
            self._TLS_WARNING_LOGGED = True
            _LOGGER.warning(
                "TLS verification turned off, its highly recommended to use a secured connection, "
                "see configuration file for configuration options"
            )

        response = requests.get(
            api_url, verify=self.tls_verify, headers={"Accept": "application/json"}
        )

        self._thoth_search_ui_url = response.headers.get("X-Thoth-Search-Ui-Url")

        try:
            response.raise_for_status()
        except Exception as exc:
            if response.status_code == 503:
                _LOGGER.error("Thoth service at %r is unavailable (HTTP 503)", api_url)
                raise ServiceUnavailable(str(exc))

            raise NoApiSupported(
                "Server does not support API v1 required by Thamos client"
            ) from exc

        self._api_url = api_url
        return self._api_url

    def check_runtime_environment(
        self, runtime_environment_name: str
    ) -> List[Dict[str, Any]]:
        """Check the given runtime environment entry."""
        runtime_environment = self.get_runtime_environment(runtime_environment_name)

        result = []

        # CUDA
        cuda_version = discover_cuda_version()
        conf_cuda_version = runtime_environment.get("cuda_version")
        if conf_cuda_version != cuda_version:
            if (
                cuda_version is None
                or isinstance(cuda_version, str)
                and isinstance(conf_cuda_version, str)
            ):
                message_type = "ERROR"
            else:
                message_type = "WARNING"

            result.append(
                {
                    "type": message_type,
                    "runtime_environment": runtime_environment_name,
                    "message": f"CUDA version declared in the configuration file "
                    f"({conf_cuda_version!r}) does not match the one detected ({cuda_version!r})",
                }
            )

        # Operating system
        conf_operating_system = runtime_environment.get("operating_system")
        if conf_operating_system:
            conf_os_name = map_os_name(conf_operating_system.get("name"))
            conf_os_version = normalize_os_version(
                conf_os_name, conf_operating_system.get("version")
            )

            if conf_os_name == "ubi":
                result.append(
                    {
                        "type": "INFO",
                        "runtime_environment": runtime_environment_name,
                        "message": "UBI container images are ABI compatible with RHEL container images",
                    }
                )
                conf_os_name = "rhel"

            os_name, os_version = discover_distribution()
            os_name = map_os_name(os_name)
            os_version = normalize_os_version(os_name, os_version)
            if conf_os_name != os_name:
                result.append(
                    {
                        "type": "ERROR",
                        "runtime_environment": runtime_environment_name,
                        "message": f"Operating system name stated in the configuration file ({conf_os_name!r}) "
                        f"does not match the one detected ({os_name!r})",
                    }
                )
            elif conf_os_version != os_version:
                result.append(
                    {
                        "type": "ERROR",
                        "runtime_environment": runtime_environment_name,
                        "message": f"Operating system version stated in the configuration file ({conf_os_version!r}) "
                        f"does not match the one detected ({os_version!r})",
                    }
                )

        # Python version
        python_version = discover_python_version()
        conf_python_version = runtime_environment.get("python_version")
        if python_version != conf_python_version:
            result.append(
                {
                    "type": "ERROR",
                    "runtime_environment": runtime_environment_name,
                    "message": f"Python version detected ({python_version!r}) does not match the one stated in the "
                    f"configuration file ({conf_python_version!r})",
                }
            )

        # Check hardware
        conf_cpu_family = runtime_environment.get("hardware", {}).get("cpu_family")
        conf_cpu_model = runtime_environment.get("hardware", {}).get("cpu_model")
        cpu_info = discover_cpu()
        if cpu_info.get("cpu_family") != conf_cpu_family:
            result.append(
                {
                    "type": "ERROR" if conf_cpu_family is not None else "WARNING",
                    "runtime_environment": runtime_environment_name,
                    "message": f"CPU family stated in the configuration file ({conf_cpu_family}) does not match the "
                    f"one detected ({cpu_info.get('cpu_family')})",
                }
            )

        if cpu_info.get("cpu_model") != conf_cpu_model:
            result.append(
                {
                    "type": "ERROR" if conf_cpu_model is not None else "WARNING",
                    "runtime_environment": runtime_environment_name,
                    "message": f"CPU model stated in the configuration file ({conf_cpu_model}) does not match the "
                    f"one detected ({cpu_info.get('cpu_model')})",
                }
            )

        return result

    def check(
        self, runtime_environment_name: Optional[str] = None
    ) -> List[Dict[str, Any]]:
        """Check the configuration file and produce a report."""
        result = []

        self.check_schema()

        # warn on multiple runtime environments when no overlay is used
        if (
            len(self.content["runtime_environments"]) > 1
            and self.content.get("overlays_dir") is None
        ):
            result.append(
                {
                    "message": "Multiple runtime environments defined but no overlays directory configured",
                    "type": "WARNING",
                }
            )

        if self.content.get("overlays_dir"):
            with workdir():
                for file_type in (
                    "Pipfile",
                    "Pipfile.lock",
                    "requirements.txt",
                    "requirements.in",
                ):
                    if os.path.isfile(file_type):
                        result.append(
                            {
                                "message": f"Overlays configured but {file_type!r} file found in the repo root, "
                                f"this might lead to misleading repository interpretation",
                                "type": "WARNING",
                            }
                        )

        if runtime_environment_name is not None:
            result.extend(self.check_runtime_environment(runtime_environment_name))
        else:
            for entry in self.content.get("runtime_environments") or []:
                result.extend(self.check_runtime_environment(entry["name"]))

        return result

    def check_schema(self) -> None:
        """Check the configuration file schema."""
        try:
            validate(instance=self.content, schema=_CONFIG_SCHEMA)
        except Exception:
            _LOGGER.error(
                "Schema validation failed: please make sure you run Thamos which supports the supplied "
                "configuration file"
            )
            raise

    def get_overlays_directory(
        self,
        runtime_environment_name: Optional[str] = None,
        *,
        missing_dir_ok: bool = False,
    ) -> str:
        """Get path to an overlays directory."""
        runtime_environment_config = self.get_runtime_environment(
            runtime_environment_name
        )
        overlays_dir = self.content.get("overlays_dir")

        with workdir(self.CONFIG_NAME):
            # No overlays directory configured.
            if overlays_dir is None:
                return os.getcwd()

            runtime_environment_config_name = runtime_environment_config["name"]
            path = os.path.join(overlays_dir, runtime_environment_config_name)
            if not missing_dir_ok and not os.path.isdir(path):
                suffix = (
                    f" --runtime-environment {runtime_environment_config_name!r}"
                    if runtime_environment_name
                    else ""
                )
                raise ConfigurationError(
                    f"The directory structure for {runtime_environment_config_name!r} is not initialized yet, "
                    f"you can initialize it by adding packages using "
                    f"`thamos add <pkg>{suffix}`",
                )

            return os.path.abspath(path)

    def get_project(
        self,
        runtime_environment_name: Optional[str] = None,
        *,
        missing_dir_ok: bool = False,
    ) -> Project:
        """Get the given overlay."""
        path = self.get_overlays_directory(
            runtime_environment_name=runtime_environment_name,
            missing_dir_ok=missing_dir_ok,
        )
        runtime_environment = RuntimeEnvironment.from_dict(
            self.get_runtime_environment(runtime_environment_name)
        )
        if self.requirements_format == "pipenv":
            pipfile_lock_path: Optional[str] = os.path.join(path, "Pipfile.lock")
            if pipfile_lock_path and not os.path.exists(pipfile_lock_path):
                pipfile_lock_path = None

            pipfile_path = os.path.join(path, "Pipfile")
            if not os.path.isfile(pipfile_path):
                if not os.path.isdir(path):
                    _LOGGER.info("Creating directory structure in %r", path)
                    os.makedirs(path, exist_ok=True)
                pipfile = Pipfile.from_dict({})
                pipfile.to_file(path=pipfile_path)

            project = Project.from_files(
                pipfile_path=pipfile_path,
                pipfile_lock_path=pipfile_lock_path,
                runtime_environment=runtime_environment,
                without_pipfile_lock=pipfile_lock_path is None,
            )
        else:
            requirements_in_file_path = os.path.join(path, "requirements.in")
            if not os.path.isfile(requirements_in_file_path):
                requirements_txt_file_path = os.path.join(path, "requirements.txt")
                if os.path.isfile(requirements_txt_file_path):
                    _LOGGER.warning(
                        "Using %r for direct dependencies", requirements_in_file_path
                    )
                    project = Project.from_pip_compile_files(
                        requirements_path=requirements_txt_file_path,
                        requirements_lock_path=None,
                        allow_without_lock=True,
                        runtime_environment=runtime_environment,
                    )
                else:
                    raise NotImplementedError(
                        "No requirements.txt/requirements.in files found, it is recommended to "
                        "use Pipenv files for managing dependencies"
                    )
            else:
                project = Project.from_pip_compile_files(
                    requirements_path=requirements_in_file_path,
                    requirements_lock_path=None,
                    allow_without_lock=True,
                    runtime_environment=runtime_environment,
                )

        return project

    def save_project(self, project: Project) -> None:
        """Save the given project to disc, performs noop if project is not dirty."""
        old_project = self.get_project(project.runtime_environment.name)
        if old_project.runtime_environment != project.runtime_environment:
            self.set_runtime_environment(
                runtime_environment=project.runtime_environment.to_dict()
            )
            self.save_config()

        if old_project.pipfile != project.pipfile:
            if config.requirements_format == "pipenv":
                pipfile_path = os.path.join(
                    self.get_overlays_directory(project.runtime_environment.name)
                )
                project.pipfile.to_file(path=pipfile_path, keep_thoth_section=True)
            else:
                requirements_in_file = self.get_overlays_directory(
                    project.runtime_environment.name
                )
                with open(
                    os.path.join(requirements_in_file, "requirements.in"), "w"
                ) as f:
                    f.write(project.pipfile.to_requirements_file(develop=False))


config = _Configuration()



#!/usr/bin/env python3
# thamos
# Copyright(C) 2018 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Core parts of library for interacting with Thoth."""

import datetime
import logging
import os
import platform
import random
import sys
import typing
from itertools import chain
from time import sleep
from time import monotonic
from contextlib import contextmanager
from functools import partial
from functools import wraps
import pprint
import json
import urllib3
import requests

import yaml
import micropipenv
from termcolor import colored
from invectio import gather_library_usage
from yaspin import yaspin
from yaspin.spinners import Spinners
from thoth.analyzer import run_command
from thoth.python import Project
from thoth.python import Constraints
from thoth.common import ThothAdviserIntegrationEnum
from thoth.common import get_justification_link as jl
from thoth.common import cwd

import thamos.discover
from . import __version__ as thamos_version
from .swagger_client.rest import ApiException
from .swagger_client import ApiClient
from .swagger_client import BuildAnalysisApi
from .swagger_client import Configuration
from .swagger_client import PythonStack
from .swagger_client import AdviseInput
from .swagger_client import ProvenanceInput
from .swagger_client import AdviseApi
from .swagger_client import ImageAnalysisApi
from .swagger_client import ProvenanceApi
from .swagger_client import ContainerImagesApi
from .swagger_client import EnvironmentsApi
from .swagger_client import PythonPackagesApi
from .swagger_client.models import AnalysisResultResponse
from .config import config as thoth_config
from .discover import discover_all
from .exceptions import ApiError
from .exceptions import NoDevRequirements
from .exceptions import NoMatchingPackage
from .exceptions import NoRequirementsFile
from .exceptions import PedanticRunVerificationError
from .exceptions import TimeoutError
from .exceptions import UnknownAnalysisType

from typing import Callable, Any, Union, Dict

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

LAST_ANALYSIS_ID_FILE = ".thoth_last_analysis_id"
_RECOMMENDATION_TYPES_URL = "https://thoth-station.ninja/recommendation-types/"

_LOGGER = logging.getLogger(__name__)
_RETRY_ON_ERROR_COUNT = int(os.getenv("THAMOS_RETRY_ON_ERROR_COUNT", 3))
_RETRY_ON_ERROR_SLEEP = float(os.getenv("THAMOS_RETRY_ON_ERROR_SLEEP", 3.0))
_THAMOS_TIMEOUT = int(os.getenv("THAMOS_TIMEOUT", 2700))
_THAMOS_TOKEN = os.getenv("THAMOS_TOKEN")
_SOURCE = {
    "adviser": "advise/python",
    "provenance-checker": "provenance/python",
    "package-extract": "analyze",
}


def with_api_client(func: typing.Callable):
    """Load configuration entries from Thoth configuration file."""
    # noqa
    @wraps(func)
    def wrapper(*args, **kwargs):
        config = Configuration()
        host = thoth_config.explicit_host
        if not host:
            thoth_config.load_config()
            host = thoth_config.content.get("host") or config.host
        thoth_config.api_discovery(host)
        _LOGGER.debug("Using API: %s", thoth_config.api_url)
        config.host = thoth_config.api_url
        config.verify_ssl = thoth_config.tls_verify

        start = monotonic()
        api_client = ApiClient(configuration=config)
        # Override default user-agent.
        api_client.user_agent = (
            f"Thamos/{thamos_version} (Python {platform.python_version()}; "
            f"{platform.system()} {platform.release()})"
        )
        result = func(api_client, *args, **kwargs)
        _LOGGER.debug("Elapsed seconds processing request: %f", monotonic() - start)
        return result

    return wrapper


def is_analysis_ready(analysis_id: str, *, verify_tls: bool = True) -> bool:
    """Handle the multiple response types available while asking for result of an analysis."""
    config = Configuration()
    host = thoth_config.explicit_host
    if not host:
        thoth_config.load_config()
        host = thoth_config.content.get("host") or config.host
    source = analysis_id.rsplit("-", maxsplit=2)[0]
    source_url = _SOURCE.get(source)
    response = requests.get(
        f"https://{host}/api/v1/{source_url}/{analysis_id}", verify=verify_tls
    )
    if response.status_code == 202:
        return False
    elif response.status_code in (200, 400):
        # Return true if result is ready.
        return True

    raise ApiError(
        f"Thoth Backend didn't respond with correct status code. Returned code - "
        f"{response.status_code}: {response.text}"
    )


def _get_spinner() -> object:
    """Choose the spinner, wisely."""
    today = datetime.datetime.now()
    if today.month == 12 and today.day in (24, 25, 26):  # Christmas.
        return Spinners.christmas
    if (today.month == 1 and today.day == 1) or (
        today.month == 12 and today.day == 31
    ):  # End or start of the year.
        return Spinners.earth
    if today.month == 2 and today.day == 14:  # Valentine's day.
        return Spinners.hearts
    if today.month == 4 and today.day == 1:  # Fool's day.
        return Spinners.smiley
    if today.month == 4 and today.day == 17:  # Easter.
        return Spinners.orangeBluePulse
    if today.month == 10 and today.day == 31:  # Halloween.
        return Spinners.monkey
    if today.hour == 7:  # A morning run.
        return Spinners.runner
    if today.hour >= 22:  # Isn't it too late?
        return Spinners.moon

    return Spinners.clock if random.random() > 0.1 else Spinners.mindblown


def _wait_for_analysis(
    status_func: Callable[..., Any],
    analysis_id: str,
    *,
    timeout: typing.Optional[int] = None,
    verify_tls: bool = True,
) -> None:
    """Wait for ongoing analysis to finish."""
    # noqa
    @contextmanager
    def _no_spinner():
        yield

    spinner = partial(
        yaspin,
        _get_spinner(),
        text=f"Waiting for response from Thoth (analysis: {analysis_id})...",
    )  # type: Union[Callable[..., Any], partial[Any]]
    if _LOGGER.getEffectiveLevel() == logging.DEBUG or bool(
        int(os.getenv("THAMOS_NO_PROGRESSBAR", 1))
    ):
        # CLI automatically injects THAMOS_NO_PROGRESSBAR=0 if user did not turned it off explictily.
        spinner = _no_spinner

    sleep_time = 0.5
    retries = 0
    timeout = timeout if timeout is not None else _THAMOS_TIMEOUT
    with spinner():
        start_time = monotonic()
        while True:
            if timeout and monotonic() - start_time > timeout:
                raise TimeoutError(
                    f"Thoth backend did not respond in time, timeout set "
                    f"to {timeout} - see {jl('thamos_timeout')}"
                )

            try:
                response = status_func(analysis_id, verify_tls=verify_tls)
            except Exception as exc:
                if retries >= _RETRY_ON_ERROR_COUNT:
                    raise

                retries += 1
                _LOGGER.error("Failed to obtain status from Thoth: %s", str(exc))
                _LOGGER.warning(
                    "Retrying in a few moments... (attempt %d/%d)",
                    retries,
                    _RETRY_ON_ERROR_COUNT,
                )
                sleep(_RETRY_ON_ERROR_SLEEP)
                continue

            retries = 0  # Reset counter as we obtained a valid response.
            if response:
                break

            if _LOGGER.getEffectiveLevel() <= logging.DEBUG:
                _LOGGER.debug(
                    "Waiting for %r to finish for %g seconds (state: %s)",
                    analysis_id,
                    sleep_time,
                    (get_status(analysis_id) or {}).get("state", "N/A"),
                )

            sleep(sleep_time)
            sleep_time = min(sleep_time * 2, 8)


def _note_last_analysis_id(analysis_id: str) -> None:
    """Store analysis id in a temporary file to keep analysis id for thamos log optional."""
    if int(os.getenv("THAMOS_DISABLE_LAST_ANALYSIS_ID_FILE", 0)):
        _LOGGER.debug("Last analysis id will not be noted")
        return

    _LOGGER.debug("Noting last analysis id %r", analysis_id)
    try:
        with open(LAST_ANALYSIS_ID_FILE, "w") as analysis_id_file:
            analysis_id_file.write(analysis_id)
    except Exception as exc:
        _LOGGER.warning("Failed to write analysis id to a temporary file: %s", str(exc))


def get_last_analysis_id() -> str:
    """Retrieve last analysis id from a temporary file."""
    try:
        with open(LAST_ANALYSIS_ID_FILE, "r") as analysis_id_file:
            analysis_id = analysis_id_file.readline().strip()
    except Exception as exc:
        raise FileNotFoundError(
            "Cannot retrieve last analysis id, you need to provide analysis id explicitly"
        ) from exc

    return analysis_id


def _retrieve_analysis_result(
    retrieve_func: Callable[..., Any], analysis_id: str
) -> Any:
    """Retrieve analysis result, raise error if analysis failed."""
    retries = 0
    while True:
        try:
            return retrieve_func(analysis_id)
        except ApiException as exc:
            if exc.status == 400:
                # Re-raise if the exception is bad request (e.g. analysis did not finish successfully).
                raise

            _LOGGER.debug(
                "Retrieved error response %s from server: %s", exc.status, exc.reason
            )
            response = json.loads(exc.body)

            if "error" in response:
                # Error produced based on API endpoints semantics...
                _LOGGER.debug("Error from Thoth: %s", response)
                _LOGGER.error("%s (analysis: %s)", response["error"], analysis_id)
            else:
                # Other errors (e.g. internal server error).
                _LOGGER.error("Error from Thoth: %s", response)

            if retries >= _RETRY_ON_ERROR_COUNT:
                return None

            retries += 1
            _LOGGER.warning(
                "Retrying in a few moments... (attempt %d/%d)",
                retries,
                _RETRY_ON_ERROR_COUNT,
            )
            sleep(_RETRY_ON_ERROR_SLEEP)


def get_static_analysis(
    src_path: str = ".",
    *,
    without_standard_imports: bool = False,
    without_builtin_imports: bool = False,
    without_builtins: bool = False,
) -> typing.Optional[dict]:
    """Get static analysis of files used in project."""
    # We are running in the root directory of project, use the root part for gathering static analysis.
    _LOGGER.info("Performing static analysis of sources to gather library usage")
    try:
        library_usage = gather_library_usage(
            src_path,
            ignore_errors=True,
            without_standard_imports=without_standard_imports,
            without_builtin_imports=without_builtin_imports,
            without_builtins=without_builtins,
        )
    except FileNotFoundError:
        _LOGGER.warning("No library usage was aggregated - no Python sources found")
        return None

    report = {}  # type: Dict[Any, Any]
    for file_record in library_usage["report"].values():
        for library, usage in file_record.items():
            # We could filter out some of the libraries which were used.
            if library not in report:
                report[library] = set()

            report[library].update(usage)

    return {
        "report": {k: list(v) for k, v in report.items()},
        "version": library_usage["version"],
    }


def _is_s2i() -> bool:
    """Check if we run in an OpenShift s2i build."""
    # This environment variable is used by OpenShift's s2i build process.
    return "STI_SCRIPTS_PATH" in os.environ


def _get_origin() -> typing.Optional[str]:
    """Check git origin configured."""
    result = run_command("git config --get remote.origin.url", raise_on_error=False)
    if result.return_code != 0:
        _LOGGER.debug(
            "Failed to obtain information about git origin: %s", result.stderr
        )
        return None

    origin = result.stdout.strip()
    if origin:
        _LOGGER.debug("Found git origin %r", origin)
        return origin

    return None


def advise_using_config(
    pipfile: str,
    pipfile_lock: str,
    config: str,
    *,
    runtime_environment_name: typing.Optional[str] = None,
    constraints: typing.Optional[str] = None,
    src_path: str = ".",
    recommendation_type: str = None,
    dev: bool = False,
    no_static_analysis: bool = False,
    no_user_stack: bool = False,
    nowait: bool = False,
    force: bool = False,
    debug: bool = False,
    origin: typing.Optional[str] = None,
    timeout: typing.Optional[int] = None,
    github_event_type: typing.Optional[str] = None,
    github_check_run_id: typing.Optional[int] = None,
    github_installation_id: typing.Optional[int] = None,
    github_base_repo_url: typing.Optional[str] = None,
    source_type: typing.Optional[ThothAdviserIntegrationEnum] = None,
    labels: typing.Optional[Dict[str, str]] = None,
) -> typing.Optional[typing.Tuple[typing.Dict[str, typing.Any], bool]]:
    """Trigger advise, respecting the configuration file supplied directly as a string or as a file path."""
    try:
        thoth_config.load_config_from_file(config)
    except (FileNotFoundError, IOError):
        thoth_config.load_config_from_string(config)

    return advise(
        pipfile=pipfile,
        pipfile_lock=pipfile_lock,
        constraints=constraints,
        recommendation_type=recommendation_type,
        src_path=src_path,
        runtime_environment=thoth_config.get_runtime_environment(
            runtime_environment_name
        ),
        runtime_environment_name=runtime_environment_name,
        dev=dev,
        no_static_analysis=no_static_analysis,
        no_user_stack=no_user_stack,
        nowait=nowait,
        force=force,
        debug=debug,
        origin=origin,
        timeout=timeout,
        github_event_type=github_event_type,
        github_check_run_id=github_check_run_id,
        github_installation_id=github_installation_id,
        github_base_repo_url=github_base_repo_url,
        source_type=source_type,
        verify_tls=thoth_config.tls_verify,
        labels=labels,
    )


@with_api_client
def advise(
    api_client: ApiClient,
    pipfile: str,
    pipfile_lock: str,
    *,
    constraints: typing.Optional[str] = None,
    recommendation_type: typing.Optional[str] = None,
    runtime_environment: typing.Optional[dict] = None,
    src_path: str = ".",
    runtime_environment_name: typing.Optional[str] = None,
    dev: bool = False,
    no_static_analysis: bool = False,
    no_user_stack: bool = False,
    nowait: bool = False,
    force: bool = False,
    debug: bool = False,
    origin: typing.Optional[str] = None,
    timeout: typing.Optional[int] = None,
    github_event_type: typing.Optional[str] = None,
    github_check_run_id: typing.Optional[int] = None,
    github_installation_id: typing.Optional[int] = None,
    github_base_repo_url: typing.Optional[str] = None,
    source_type: typing.Optional[ThothAdviserIntegrationEnum] = None,
    justification: typing.Optional[Dict] = None,
    stack_info: typing.Optional[Dict] = None,
    kebechet_metadata: typing.Optional[Dict] = None,
    verify_tls: bool = True,
    labels: typing.Optional[Dict[str, str]] = None,
) -> typing.Optional[tuple]:
    """Submit a stack for adviser checks and wait for results."""
    if not pipfile:
        raise ValueError("No Pipfile content provided for advises")

    if runtime_environment is None:
        runtime_environment = thoth_config.get_runtime_environment(
            runtime_environment_name
        )

    if runtime_environment_name is None and runtime_environment:
        runtime_environment_name = runtime_environment.get("name")

    if (runtime_environment and runtime_environment_name) and runtime_environment.get(
        "name"
    ) != runtime_environment_name:
        raise ValueError(
            "Runtime environment name supplied does not match with the one stated in the config file"
        )

    if recommendation_type is None:
        priority = (
            runtime_environment.pop("recommendation_type", None),
            thoth_config.content.get("recommendation_type", None),
            "stable",
        )
        recommendation_type = next(filter(bool, priority))

    _LOGGER.info(
        "Using %r recommendation type - see %s",
        recommendation_type,
        _RECOMMENDATION_TYPES_URL,
    )

    if no_user_stack and pipfile_lock:
        _LOGGER.warning(
            "The user stack found in the lock file will not be supplied as requested"
        )
        pipfile_lock = ""

    library_usage = None
    if not no_static_analysis:
        library_usage = get_static_analysis(src_path)
        _LOGGER.debug(
            "Library usage:%s",
            "\n" + json.dumps(library_usage, indent=2) if library_usage else None,
        )

    stack = PythonStack(requirements=pipfile, requirements_lock=pipfile_lock or "")

    if runtime_environment:
        # Override recommendation type specified explicitly in the runtime environment entry.
        runtime_environment.pop("recommendation_type", None)

    input_args: Dict[str, Any] = {
        "application_stack": stack,
        "runtime_environment": runtime_environment,
        "library_usage": library_usage,
    }

    if kebechet_metadata:
        input_args["kebechet_metadata"] = kebechet_metadata
    if justification:
        input_args["justification"] = justification
    if stack_info:
        input_args["stack_info"] = stack_info
    if constraints:
        input_args["constraints"] = constraints
    if labels:
        input_args["labels"] = labels
    elif runtime_environment.get("labels"):
        input_args["labels"] = runtime_environment["labels"]

    advise_input = AdviseInput(**input_args)
    api_instance = AdviseApi(api_client)

    if recommendation_type:
        recommendation_type = recommendation_type.lower()

    if source_type is None and _is_s2i():
        source_type = ThothAdviserIntegrationEnum.S2I

    parameters = {
        "recommendation_type": recommendation_type,
        "debug": debug,
        "force": force,
        "source_type": source_type.name.lower() if source_type is not None else None,
        "origin": _get_origin(),
        "dev": dev,
    }  # type: Dict[str, Any]

    if origin is not None:
        parameters["origin"] = origin

    if github_event_type is not None:
        parameters["github_event_type"] = github_event_type

    if github_check_run_id is not None:
        parameters["github_check_run_id"] = github_check_run_id

    if github_installation_id is not None:
        parameters["github_installation_id"] = github_installation_id

    if github_base_repo_url is not None:
        parameters["github_base_repo_url"] = github_base_repo_url

    if _THAMOS_TOKEN:
        parameters["token"] = _THAMOS_TOKEN

    response = api_instance.post_advise_python(advise_input, **parameters)

    _LOGGER.info(
        "Successfully submitted advise analysis %r to %r",
        response.analysis_id,
        thoth_config.api_url,
    )

    _note_last_analysis_id(response.analysis_id)

    _LOGGER.debug("Analysis parameters:\n%r", pprint.pformat(parameters))
    _LOGGER.debug("Adviser input:\n%s", advise_input.to_str())

    if nowait:
        return response.analysis_id
    # We call custom status function for advise until swagger client supports mulitple response codes.
    _wait_for_analysis(
        is_analysis_ready, response.analysis_id, timeout=timeout, verify_tls=verify_tls
    )
    _LOGGER.debug("Retrieving adviser result for %r", response.analysis_id)
    response = _retrieve_analysis_result(
        api_instance.get_advise_python, response.analysis_id
    )
    if not response:
        return None

    _LOGGER.debug("Adviser check metadata: %r", response.metadata)
    return response.result.to_dict(), response.result.error


def advise_here(
    recommendation_type: typing.Optional[str] = None,
    *,
    runtime_environment: typing.Optional[dict] = None,
    src_path: str = ".",
    runtime_environment_name: typing.Optional[str] = None,
    dev: bool = False,
    no_static_analysis: bool = False,
    no_user_stack: bool = False,
    nowait: bool = False,
    force: bool = False,
    debug: bool = False,
    timeout: typing.Optional[int] = None,
    origin: typing.Optional[str] = None,
    github_event_type: typing.Optional[str] = None,
    github_check_run_id: typing.Optional[int] = None,
    github_installation_id: typing.Optional[int] = None,
    github_base_repo_url: typing.Optional[str] = None,
    source_type: typing.Optional[ThothAdviserIntegrationEnum] = None,
    justification: typing.Optional[Dict] = None,
    stack_info: typing.Optional[Dict] = None,
    kebechet_metadata: typing.Optional[Dict] = None,
    verify_tls: bool = True,
    labels: typing.Optional[Dict[str, str]] = None,
) -> typing.Optional[tuple]:
    """Run advise in current directory, requires no arguments."""
    if runtime_environment_name is not None and runtime_environment is not None:
        raise ValueError(
            "Both runtime_environment and runtime_environment_name were set."
        )

    if runtime_environment is None and runtime_environment_name is None:
        runtime_environment = thoth_config.get_runtime_environment()

    if runtime_environment:
        runtime_environment_name = runtime_environment["name"]

    # get_overlays_directory() returns the current working directory if overlays_dir is not configured so this change
    # should not result in a behavioral change
    overlay_dir = thoth_config.get_overlays_directory(
        runtime_environment_name=runtime_environment_name
    )

    with cwd(overlay_dir):
        requirements_format = thoth_config.requirements_format
        if requirements_format == "pipenv":
            _LOGGER.info(
                "Using Pipenv files to manage dependencies located in %r", os.getcwd()
            )
            pipfile_lock_exists = os.path.exists("Pipfile.lock")

            if pipfile_lock_exists:
                _LOGGER.info(
                    "Submitting Pipfile.lock as a base for user's stack scoring - see %s",
                    jl("user_stack"),
                )

            project = Project.from_files(
                pipfile_path="Pipfile",
                pipfile_lock_path="Pipfile.lock",
                without_pipfile_lock=not os.path.exists("Pipfile.lock"),
            )

            if (
                pipfile_lock_exists
                and project.pipfile_lock.meta.hash["sha256"]
                != project.pipfile.hash()["sha256"]
            ):
                _LOGGER.error(
                    "Pipfile hash stated in Pipfile.lock %r does not correspond to Pipfile hash %r - was Pipfile "
                    "adjusted? This error is not critical.",
                    project.pipfile_lock.meta.hash["sha256"][:6],
                    project.pipfile.hash()["sha256"][:6],
                )
        elif requirements_format in ("pip", "pip-tools", "pip-compile"):
            _LOGGER.info(
                "Using requirements.txt file to manage dependencies located in %r",
                os.getcwd(),
            )
            project = Project.from_pip_compile_files(allow_without_lock=True)
        else:
            raise ValueError(
                f"Unknown configuration option for requirements format: {requirements_format!r}"
            )

        constraints_str = None
        if os.path.exists("constraints.txt"):
            _LOGGER.info("Using constraints.txt file located in %r", os.getcwd())
            with open("constraints.txt") as constraints_file:
                constraints_str = constraints_file.read()

            # Try to load constraints before the request to verify their correctness without sending request to the
            # backend.
            Constraints.from_string(constraints_str)

        pipfile = project.pipfile.to_string()
        pipfile_lock_str = (
            project.pipfile_lock.to_string() if project.pipfile_lock else ""
        )

        return advise(
            pipfile=pipfile,
            pipfile_lock=pipfile_lock_str,
            constraints=constraints_str,
            recommendation_type=recommendation_type,
            src_path=src_path,
            runtime_environment=runtime_environment,
            runtime_environment_name=runtime_environment_name,
            dev=dev,
            no_static_analysis=no_static_analysis,
            no_user_stack=no_user_stack,
            nowait=nowait,
            force=force,
            debug=debug,
            origin=origin,
            timeout=timeout,
            source_type=source_type,
            github_event_type=github_event_type,
            github_check_run_id=github_check_run_id,
            github_installation_id=github_installation_id,
            github_base_repo_url=github_base_repo_url,
            justification=justification,
            stack_info=stack_info,
            kebechet_metadata=kebechet_metadata,
            verify_tls=verify_tls,
            labels=labels,
        )


@with_api_client
def provenance_check(
    api_client: ApiClient,
    pipfile: str,
    pipfile_lock: str,
    *,
    nowait: bool = False,
    force: bool = False,
    debug: bool = False,
    origin: typing.Optional[str] = None,
    timeout: typing.Optional[int] = None,
    justification: typing.Optional[Dict[str, Any]] = None,
    stack_info: typing.Optional[Dict[str, Any]] = None,
    kebechet_metadata: typing.Optional[Dict[str, Any]] = None,
    verify_tls: bool = False,
) -> typing.Optional[tuple]:
    """Submit a stack for provenance checks and wait for results."""
    if not pipfile:
        raise ValueError("No Pipfile content provided for provenance checks")

    input_args: Dict[str, Any] = {
        "application_stack": PythonStack(
            requirements=pipfile, requirements_lock=pipfile_lock
        ),
    }
    if justification:
        input_args["justification"] = justification
    if stack_info:
        input_args["stack_info"] = stack_info
    if kebechet_metadata:
        input_args["kebechet_metadata"] = kebechet_metadata

    provenance_input = ProvenanceInput(**input_args)

    provenance_kwargs: Dict[str, Any] = {
        "debug": debug,
        "force": force,
    }
    if origin:
        provenance_kwargs["origin"] = origin
    if _THAMOS_TOKEN:
        provenance_kwargs["token"] = _THAMOS_TOKEN

    api_instance = ProvenanceApi(api_client)
    response = api_instance.post_provenance_python(
        provenance_input, **provenance_kwargs
    )
    _LOGGER.info(
        "Successfully submitted provenance check analysis %r to %r",
        response.analysis_id,
        thoth_config.api_url,
    )

    _note_last_analysis_id(response.analysis_id)

    if nowait:
        return response.analysis_id

    _wait_for_analysis(
        is_analysis_ready, response.analysis_id, timeout=timeout, verify_tls=verify_tls
    )
    _LOGGER.debug("Retrieving provenance check result for %r", response.analysis_id)
    response = _retrieve_analysis_result(
        api_instance.get_provenance_python, response.analysis_id
    )
    if not response:
        return None

    _LOGGER.debug("Provenance check metadata: %r", response.metadata)
    result = response.result.to_dict()
    return result["report"], result["error"]


def provenance_check_here(
    *,
    nowait: bool = False,
    force: bool = False,
    debug: bool = False,
    origin: typing.Optional[str] = None,
    timeout: typing.Optional[int] = None,
    justification: typing.Optional[Dict[str, Any]] = None,
    stack_info: typing.Optional[Dict[str, Any]] = None,
    kebechet_metadata: typing.Optional[Dict[str, Any]] = None,
    verify_tls: bool = True,
) -> typing.Optional[tuple]:
    """Submit a provenance check in current directory."""
    if not os.path.isfile("Pipfile"):
        raise FileNotFoundError("No Pipfile found in current directory")

    if not os.path.isfile("Pipfile.lock"):
        raise FileNotFoundError("No Pipfile.lock found in current directory")

    with open("Pipfile", "r") as pipfile, open("Pipfile.lock", "r") as piplock:
        return provenance_check(
            pipfile.read(),
            piplock.read(),
            nowait=nowait,
            force=force,
            debug=debug,
            origin=origin,
            timeout=timeout,
            justification=justification,
            stack_info=stack_info,
            kebechet_metadata=kebechet_metadata,
            verify_tls=verify_tls,
        )


@with_api_client
def image_analysis(
    api_client: ApiClient,
    image: str,
    *,
    environment_type: str,
    registry_user: str = None,
    registry_password: str = None,
    verify_tls: bool = True,
    nowait: bool = False,
    force: bool = False,
    timeout: typing.Optional[int] = None,
    debug: bool = False,
) -> Union[Dict, str, None]:
    """Submit an image for analysis to Thoth."""
    if not image:
        raise ValueError("No image provided")

    api_instance = ImageAnalysisApi(api_client)
    if registry_user or registry_password:
        # Swagger client handles None in a different way - we need to explicitly avoid passing
        # registry user and registry password if they are not set.
        response = api_instance.post_analyze(
            image=image,
            debug=debug,
            registry_user=registry_user,
            registry_password=registry_password,
            verify_tls=verify_tls,
            force=force,
            environment_type=environment_type,
        )
    else:
        response = api_instance.post_analyze(
            image=image, debug=debug, verify_tls=verify_tls, force=force
        )

    _LOGGER.info(
        "Successfully submitted provenance check analysis %r to %r",
        response.analysis_id,
        thoth_config.api_url,
    )
    if nowait:
        return response.analysis_id

    _wait_for_analysis(
        is_analysis_ready, response.analysis_id, timeout=timeout, verify_tls=verify_tls
    )
    _LOGGER.debug(
        "Retrieving image analysis result result for %r", response.analysis_id
    )
    response = _retrieve_analysis_result(api_instance.get_analyze, response.analysis_id)
    if not response:
        return None

    _LOGGER.debug("Image analysis metadata: %r", response.metadata)
    return response.result


@with_api_client
def build_analysis(
    api_client: ApiClient,
    build_log: dict,
    base_image: str,
    output_image: str,
    *,
    environment_type: str,
    base_registry_user: typing.Optional[str] = None,
    base_registry_password: typing.Optional[str] = None,
    base_registry_verify_tls: bool = True,
    output_registry_user: typing.Optional[str] = None,
    output_registry_password: typing.Optional[str] = None,
    output_registry_verify_tls: bool = True,
    origin: typing.Optional[str] = None,
    nowait: bool = False,
    force: bool = False,
    debug: bool = False,
) -> typing.Union[typing.Dict, str]:
    """Submit a build image and logs for analysis to Thoth."""
    if build_log or base_image or output_image:
        build_detail = {
            "build_log": build_log,
            "base_image": base_image,
            "output_image": output_image,
        }
    else:
        raise ValueError("No build info provided")

    params = dict(
        body=build_detail,
        base_registry_user=base_registry_user,
        base_registry_password=base_registry_password,
        base_registry_verify_tls=base_registry_verify_tls,
        output_registry_user=output_registry_user,
        output_registry_password=output_registry_password,
        output_registry_verify_tls=output_registry_verify_tls,
        environment_type=environment_type,
        origin=origin,
        debug=debug,
        force=force,
    )

    # Swagger client handles None in a different way - we need to explicitly avoid passing
    # values that are not present.
    params = {k: v for k, v in params.items() if v is not None}

    api_instance = BuildAnalysisApi(api_client)
    response = api_instance.post_build(**params)

    _LOGGER.info("Successfully submitted build analysis to %r", thoth_config.api_url)
    _LOGGER.debug("Build analysis parameters: %r", params)
    if nowait:
        return response

    return response


@with_api_client
def get_log(api_client: ApiClient, analysis_id: str = None):
    """Get log of an analysis - the analysis type and endpoint are automatically derived from analysis id.

    If analysis_id is not provided, its get from the last thamos call which stores it in a temporary file.
    """
    if not analysis_id:
        analysis_id = get_last_analysis_id()

    if analysis_id.startswith("package-extract-"):
        api_instance = ImageAnalysisApi(
            api_client
        )  # type: Union[ImageAnalysisApi, ProvenanceApi, AdviseApi]
        method = api_instance.get_analyze_log  # type: ignore
    elif analysis_id.startswith("provenance-checker-"):
        api_instance = ProvenanceApi(api_client)
        method = api_instance.get_provenance_python_log  # type: ignore
    elif analysis_id.startswith("adviser-"):
        api_instance = AdviseApi(api_client)
        method = api_instance.get_advise_python_log  # type: ignore
    else:
        raise UnknownAnalysisType(
            "Cannot determine analysis type from identifier: %r", analysis_id
        )

    json_log = method(analysis_id).log

    if not json_log:
        return json_log

    result = ""
    for line in json_log.splitlines():
        try:
            content = json.loads(line)
            if not int(os.getenv("THAMOS_NO_EMOJI", 0)):
                if content["levelname"] == "DEBUG":
                    message = colored(content["message"], "green")
                elif content["levelname"] == "INFO":
                    message = colored(content["message"], "cyan")
                elif content["levelname"] == "WARNING":
                    message = colored(content["message"], "yellow", attrs=["bold"])
                elif content["levelname"] in ("ERROR", "CRITICAL"):
                    message = colored(content["message"], "red", attrs=["bold"])
                else:
                    message = content["message"]

                result += "{} {:<27} {}: {}\n".format(
                    content["asctime"], content["name"], content["levelname"], message
                )
            else:
                result += "{} {:<27} {}: {}\n".format(
                    content["asctime"],
                    content["name"],
                    content["levelname"],
                    content["message"],
                )
        except Exception:
            # If the content parsed does not carry logger information or has not relevant
            # entries, log the original message.
            result += line + "\n"
            continue

    return result


@with_api_client
def get_status(api_client: ApiClient, analysis_id: typing.Optional[str] = None):
    """Get status of an analysis - the analysis type and endpoint are automatically derived from analysis id.

    If analysis_id is not provided, its get from the last thamos call which stores it in a temporary file.
    """
    if not analysis_id:
        analysis_id = get_last_analysis_id()

    if analysis_id.startswith("package-extract-"):
        api_instance = ImageAnalysisApi(
            api_client
        )  # type: Union[ImageAnalysisApi, ProvenanceApi, AdviseApi]
        method = api_instance.get_analyze_status  # type: ignore
    elif analysis_id.startswith("provenance-checker-"):
        api_instance = ProvenanceApi(api_client)
        method = api_instance.get_provenance_python_status  # type: ignore
    elif analysis_id.startswith("adviser-"):
        api_instance = AdviseApi(api_client)
        method = api_instance.get_advise_python_status  # type: ignore
    else:
        raise UnknownAnalysisType(
            "Cannot determine analysis type from identifier: %r", analysis_id
        )

    return method(analysis_id).status.to_dict()


@with_api_client
def get_analysis_results(
    api_client: ApiClient, analysis_id: typing.Optional[str] = None
):
    """Get the analysis result from a given id."""
    if not analysis_id:
        analysis_id = get_last_analysis_id()

    if analysis_id.startswith("package-extract-"):
        api_instance = ImageAnalysisApi(
            api_client
        )  # type: Union[ImageAnalysisApi, ProvenanceApi, AdviseApi]
        method = api_instance.get_analyze  # type: ignore
        response = _retrieve_analysis_result(
            method, analysis_id
        )  # type: AnalysisResultResponse
        return response.result
    elif analysis_id.startswith("provenance-checker-"):
        api_instance = ProvenanceApi(api_client)
        method = api_instance.get_provenance_python  # type: ignore
        response = _retrieve_analysis_result(method, analysis_id)
        return response.result["report"], response.result["error"]
    elif analysis_id.startswith("adviser-"):
        api_instance = AdviseApi(api_client)
        method = api_instance.get_advise_python  # type: ignore
        response = _retrieve_analysis_result(method, analysis_id)
        return response.result.to_dict(), response.result.error
    else:
        raise UnknownAnalysisType(
            "Cannot determine analysis type from identifier: %r", analysis_id
        )


def install_using_config(
    config: str,
    runtime_environment_name: typing.Optional[str] = None,
    dev: bool = False,
) -> None:
    """Perform installation given the configuration supplied."""
    try:
        thoth_config.load_config_from_file(config)
    except (FileNotFoundError, IOError):
        thoth_config.load_config_from_string(config)

    install(runtime_environment_name=runtime_environment_name, dev=dev)


def install(
    runtime_environment_name: typing.Optional[str] = None,
    dev: bool = False,
    pip_args: typing.Optional[typing.Tuple[str]] = None,
) -> None:
    """Perform installation of packages for the given runtime environment.

    If the runtime environment is not specified, the first environment stated in the configuration is used.
    """
    method = (
        "pipenv" if thoth_config.requirements_format == "pipenv" else "requirements"
    )

    if not dev and method == "pipenv":
        _LOGGER.warning(
            "Development dependencies will not be installed - see %s", jl("no_dev")
        )

    with cwd(thoth_config.get_overlays_directory(runtime_environment_name)):
        if method == "pipenv":
            if not os.path.isfile("Pipfile.lock"):
                raise NoRequirementsFile(
                    f"No Pipfile.lock found in {os.getcwd()!r} needed to install dependencies, "
                    "issue `thamos advise` resolve dependencies"
                )
            if not os.path.isfile("Pipfile"):  # Required for computing digests.
                raise NoRequirementsFile(
                    f"No Pipfile found in {os.getcwd()!r} needed for the installation process"
                )

            if dev:
                with open("Pipfile.lock") as pipfile_lock_file:
                    content = json.load(pipfile_lock_file)

                if not content.get("develop"):
                    raise NoDevRequirements(
                        "No development requirements found in the lock file, make sure development "
                        "requirements are stated and the resolved stack preserves them by "
                        "using `thamos advise --dev`"
                    )

                del content
        else:
            if not os.path.isfile("requirements.txt"):
                raise NoRequirementsFile(
                    f"No requirements.txt file found in {os.getcwd()!r} needed to install dependencies"
                )

        if runtime_environment_name is None:
            config_entry = thoth_config.get_runtime_environment(
                runtime_environment_name
            )
            runtime_environment_name = config_entry["name"]
        _LOGGER.info(
            "Installing requirements for runtime environment %r",
            runtime_environment_name,
        )

        _LOGGER.info(
            "Using %r installation method to install dependencies stated in %r",
            method,
            os.getcwd(),
        )

        micropipenv_kwargs = {
            "method": method,
            "deploy": True,
            "dev": dev,
            "pip_args": pip_args,
        }

        virtualenv_path = thoth_config.get_virtualenv_path(runtime_environment_name)
        if virtualenv_path:
            if not os.path.isdir(virtualenv_path):
                thoth_config.create_virtualenv()

            micropipenv_kwargs["pip_bin"] = os.path.join(virtualenv_path, "bin", "pip3")

        # micropipenv writes and prints the lockfile which is not very user friendly when thamos is used by a user.
        # Suppress this behavior unless these environment variables options are supplied explicitly.
        old_write = os.getenv("MICROPIPENV_NO_LOCKFILE_WRITE")
        if old_write is None:
            os.environ["MICROPIPENV_NO_LOCKFILE_WRITE"] = "1"

        old_print = os.getenv("MICROPIPENV_NO_LOCKFILE_PRINT")
        if old_print is None:
            os.environ["MICROPIPENV_NO_LOCKFILE_PRINT"] = "1"

        try:
            micropipenv.install(**micropipenv_kwargs)
        finally:
            if old_write is None:
                os.environ.pop("MICROPIPENV_NO_LOCKFILE_WRITE", None)

            if old_print is None:
                os.environ.pop("MICROPIPENV_NO_LOCKFILE_PRINT", None)


@with_api_client
def list_thoth_container_images(
    api_client: ApiClient,
    *,
    os_name: typing.Optional[str] = None,
    os_version: typing.Optional[str] = None,
    python_version: typing.Optional[str] = None,
    cuda_version: typing.Optional[str] = None,
    image_name: typing.Optional[str] = None,
    library_name: typing.Optional[str] = None,
    symbol: typing.Optional[str] = None,
    package_name: typing.Optional[str] = None,
    rpm_package_name: typing.Optional[str] = None,
) -> typing.List[typing.Dict[str, Any]]:
    """Get available Thoth container images."""
    filtering = locals()
    filtering.pop("api_client", None)
    filtering = {k: v for k, v in filtering.items() if v is not None}

    result = []
    page = 0
    while True:
        images = (
            ContainerImagesApi(api_client)
            .list_thoth_container_images(page=page, **filtering)
            .to_dict()["container_images"]
        )
        if not images:
            break
        result.extend(images)
        page += 1

    return result


@with_api_client
def list_python_environments(api_client: ApiClient) -> typing.Dict[str, Any]:
    """Get available Python environments."""
    return (
        EnvironmentsApi(api_client).list_python_environments().to_dict()["environment"]
    )


@with_api_client
def list_python_package_indexes(api_client: ApiClient) -> typing.Dict[str, Any]:
    """Get information about hardware for which Thoth can give recommendations."""
    return (
        PythonPackagesApi(api_client).list_python_package_indexes().to_dict()["indexes"]
    )


@with_api_client
def get_package_from_imported_packages(
    api_client: ApiClient,
    import_name: str,
    raise_on_error: bool = True,
) -> typing.List[Dict[str, Any]]:
    """Get all (package_name, package_version, index_url) triplets for given import package name."""
    try:
        return (
            PythonPackagesApi(api_client)
            .get_package_from_imported_packages(import_name)
            .to_dict()["package_names"]
        )
    except ApiException as exc:
        error_message = f"Failed to obtain package for import {import_name!r} (HTTP status {exc.status})"
        if exc.body:
            try:
                error_message += (
                    f": {str(json.loads(exc.body.decode('utf-8'))['error'])}"
                )
            except json.decoder.JSONDecodeError:
                _LOGGER.error(
                    "Failed to decode backend response (HTTP status %r)", exc.status
                )

        if raise_on_error:
            raise NoMatchingPackage(error_message) from exc

        return []


def get_verified_packages_from_static_analysis(
    src_path: str = ".",
    *,
    without_standard_imports: bool = False,
    without_builtin_imports: bool = False,
    without_builtins: bool = False,
    raise_on_error: bool = True,
) -> typing.List[typing.Dict[str, str]]:
    """Get verified packages from invectio static analysis result."""
    # 1. Obtain list of imports using invectio
    result = get_static_analysis(
        src_path,
        without_standard_imports=without_standard_imports,
        without_builtin_imports=without_builtin_imports,
        without_builtins=without_builtins,
    )

    packages = []
    if result:
        packages = [p for p in result["report"].keys()]

    if not packages:
        _LOGGER.error("No package imports identified in %r", src_path)
        return []

    # 2. For each import verify package (name, version, index) (whatprovides logic)
    verified_packages = []

    for import_name in packages:
        unique_packages: typing.List[typing.Dict] = []

        imported_packages = get_package_from_imported_packages(
            import_name, raise_on_error=raise_on_error
        )
        if not imported_packages:
            _LOGGER.error("No package for import %r found", import_name)
            continue

        for package in imported_packages:
            if package["package_name"] not in [
                p["package_name"] for p in unique_packages
            ]:
                unique_packages.append(
                    {
                        "package_name": package["package_name"],
                        "index_url": package["index_url"],
                    }
                )
            else:
                existing_indexes = [
                    p["index_url"]
                    for p in unique_packages
                    if p["package_name"] == package["package_name"]
                ]

                if package["index_url"] not in existing_indexes:
                    unique_packages.append(
                        {
                            "package_name": package["package_name"],
                            "index_url": package["index_url"],
                        }
                    )

        for unique_package in unique_packages:
            verified_packages.append(unique_package)
            _LOGGER.info(
                "Package name %r identified for import %r",
                unique_package["package_name"],
                import_name,
            )

    return verified_packages


def write_files(
    requirements: typing.Dict[str, Any],
    requirements_lock: typing.Dict[str, Any],
    requirements_format: str,
) -> None:
    """Write content of Pipfile/Pipfile.lock or requirements.in/txt to the current directory."""
    project = Project.from_dict(requirements, requirements_lock)
    if requirements_format == "pipenv":
        _LOGGER.debug("Writing to Pipfile/Pipfile.lock in %r", os.getcwd())
        project.to_files(keep_thoth_section=True)
    elif requirements_format in ("pip", "pip-tools", "pip-compile"):
        _LOGGER.debug("Writing to requirements.in/requirements.txt in %r", os.getcwd())
        project.to_pip_compile_files()
        _LOGGER.debug("No changes to Pipfile to write")
    else:
        raise ValueError(
            f"Unknown requirements format, supported are 'pipenv' and 'pip': {requirements_format!r}"
        )


def load_files(requirements_format: str) -> typing.Tuple[str, typing.Optional[str]]:
    """Load Pipfile/Pipfile.lock or requirements.in/txt from the current directory."""
    if requirements_format == "pipenv":
        _LOGGER.info("Using Pipenv files located in %r directory", os.getcwd())
        pipfile_lock_exists = os.path.exists("Pipfile.lock")

        if pipfile_lock_exists:
            _LOGGER.info(
                "Submitting Pipfile.lock as a base for user's stack scoring - see %s",
                jl("user_stack"),
            )

        project = Project.from_files(
            pipfile_path="Pipfile",
            without_pipfile_lock=not os.path.exists("Pipfile.lock"),
        )

        if (
            pipfile_lock_exists
            and project.pipfile_lock.meta.hash["sha256"]
            != project.pipfile.hash()["sha256"]
        ):
            _LOGGER.error(
                "Pipfile hash stated in Pipfile.lock %r does not correspond to Pipfile hash %r - was Pipfile "
                "adjusted? This error is not critical.",
                project.pipfile_lock.meta.hash["sha256"][:6],
                project.pipfile.hash()["sha256"][:6],
            )
    elif requirements_format in ("pip", "pip-tools", "pip-compile"):
        _LOGGER.info("Using requirements.txt file located in %r directory", os.getcwd())
        project = Project.from_pip_compile_files(allow_without_lock=True)
    else:
        raise ValueError(
            f"Unknown configuration option for requirements format: {requirements_format!r}"
        )
    return (
        project.pipfile.to_string(),
        project.pipfile_lock.to_string() if project.pipfile_lock else None,
    )


def write_configuration(
    advised_configuration: dict,
    recommendation_type: str = None,
    dev: bool = False,
) -> None:
    """Create thoth configuration file."""
    if not advised_configuration:
        _LOGGER.debug("No advises on configuration, nothing to adjust")
        return

    if "name" not in advised_configuration:
        _LOGGER.error(
            "Cannot adjust Thoth's configuration based on advises: No name found in Thoth's configuration entry"
        )
        return

    _LOGGER.debug("Reading Thoth's configuration file")
    with open(".thoth.yaml", "r") as thoth_yaml_file:
        content = yaml.safe_load(thoth_yaml_file.read())

    for idx, runtime_environment_entry in enumerate(
        content.get("runtime_environments", [])
    ):
        if runtime_environment_entry.get("name") == advised_configuration["name"]:
            _LOGGER.debug(
                "Adjusting configuration entry for %r based on recommendations",
                advised_configuration["name"],
            )
            runtime_environment_entry = advised_configuration
            if recommendation_type:
                runtime_environment_entry["recommendation_type"] = recommendation_type
            runtime_environment_entry["dev"] = dev
            content["runtime_environments"][idx] = runtime_environment_entry
            break
    else:
        _LOGGER.error(
            "Cannot adjust Thoth's configuration based on advises: No runtime environment entry with name %r found",
            advised_configuration["name"],
        )
        return

    _LOGGER.debug("Writing adjusted Thoth's configuration file")
    with open(".thoth.yaml", "w") as thoth_yaml_file:
        yaml.safe_dump(content, thoth_yaml_file)


def collect_support_information_dict() -> Dict[str, Any]:
    """Collect environment information suitable to report bugs or issues in a dictionary form."""
    discovery = {}

    for name in dir(thamos.discover):
        if not name.startswith("discover_"):
            continue

        func = getattr(thamos.discover, name)
        discovery[name[len("discover_") :]] = func()

    thoth_config_content = None
    try:
        thoth_config_content = thoth_config.content
    except Exception:
        _LOGGER.exception("Failed to obtain content of Thoth's configuration file")

    thoth_version = None
    try:
        thoth_version = thoth_config.get_thoth_version()
    except Exception:
        _LOGGER.exception("Failed to obtain Thoth version information")

    pip_freeze = None
    try:
        pip_freeze = run_command("pip freeze").stdout.splitlines()
    except Exception:
        _LOGGER.exception("Failed to obtain packages present in the environment")

    last_analysis_id = None
    try:
        last_analysis_id = get_last_analysis_id()
    except FileNotFoundError:
        _LOGGER.warning("Cannot retrieve last analysis identifier")
        pass

    result = {
        "environment": {
            k: v for k, v in os.environ.items() if k.startswith(("THAMOS_", "THOTH_"))
        },
        "is_s2i": _is_s2i(),
        "last_analysis_id": last_analysis_id,
        "thoth_config": thoth_config_content,
        "thoth_version": thoth_version,
        "thamos_version": thamos_version,
        "os_name": os.name,
        "platform_machine": platform.machine(),
        "platform_python_implementation": platform.python_implementation(),
        "platform_release": platform.release(),
        "platform_system": platform.system(),
        "platform_version": platform.version(),
        "python_full_version": platform.python_version(),
        "python_version": ".".join(platform.python_version().split(".")[:2]),
        "sys_platform": sys.platform,
        "pip_freeze": pip_freeze,
        "discovery": discovery,
    }
    return result


def _get_package_entry_str(
    pipfile_lock: Dict[str, Dict[str, typing.Any]], package_name: str
) -> str:
    """Get information about the given package so that they are printed to users."""
    package_entry = pipfile_lock["default"].get(
        package_name, pipfile_lock["develop"].get(package_name)
    )
    if not package_entry:
        # Any error spotted?
        return "UNKNOWN==UNKNOWN from UNKNOWN"

    index_url = "UNKNOWN"
    for source in pipfile_lock["_meta"]["sources"]:
        if package_entry.get("index") == source["name"]:
            index_url = source["url"]
            break

    return f"{package_name}{package_entry.get('version')} from {index_url}"


def _traverse_edges(
    pipfile_lock: Dict[str, Dict[str, str]],
    nodes: typing.List[str],
    edges: typing.List[typing.List[int]],
    seen_nodes: typing.Set[int],
    *,
    starting_node: int,
    indent: int = 2,
    fold: bool = True,
) -> None:
    """Traverse edges of the dependency graph and show its structure to terminal."""
    for edge in edges:
        if edge[0] == starting_node:
            print(
                " " * indent
                + f"↳ {_get_package_entry_str(pipfile_lock, nodes[edge[1]])}"
            )
            if edge[1] not in seen_nodes:
                if fold:
                    seen_nodes.add(edge[1])

                _traverse_edges(
                    pipfile_lock,
                    nodes,
                    edges,
                    seen_nodes=seen_nodes | {edge[1]},
                    starting_node=edge[1],
                    indent=indent + 2,
                    fold=fold,
                )
            else:
                # Note cyclic dependency.
                print(" " * (indent + 2) + "↳ …")


def print_dependency_graph_from_adviser_document(
    adviser_document: typing.Dict[str, typing.Any], *, fold: bool = True
) -> bool:
    """Print dependency graph to stdout from the given document."""
    if not adviser_document["report"]["products"]:
        _LOGGER.error(
            "No dependency graph to show, was the adviser request successful?"
        )
        return False

    if not adviser_document["report"]["products"][0].get("dependency_graph"):
        # XXX: we can remove this check over time.
        _LOGGER.error(
            "No dependency graph to show, was the result computed with recent adviser?"
        )
        return False

    nodes = adviser_document["report"]["products"][0]["dependency_graph"]["nodes"]
    edges = adviser_document["report"]["products"][0]["dependency_graph"]["edges"]

    nodes_idx = {n: i for i, n in enumerate(nodes)}
    # Keep a list of flags stating the given package is a root dependency.
    pipfile = adviser_document["report"]["products"][0]["project"]["requirements"]
    pipfile_lock = adviser_document["report"]["products"][0]["project"][
        "requirements_locked"
    ]

    direct_dependencies = pipfile["packages"].keys()
    if adviser_document["parameters"]["dev"]:
        # Include dev packages in the listing only if dev flag was supplied
        direct_dependencies = chain(direct_dependencies, pipfile["dev-packages"].keys())
    else:
        _LOGGER.warning(
            "Development dependencies will not be shown as no --dev flag was supplied to the resolution process"
        )

    for direct_dependency in direct_dependencies:
        direct_dependency_idx = nodes_idx[direct_dependency]
        print(f"→ {_get_package_entry_str(pipfile_lock, direct_dependency)}")
        _traverse_edges(
            pipfile_lock,
            nodes,
            edges,
            seen_nodes={direct_dependency_idx},
            starting_node=direct_dependency_idx,
            fold=fold,
        )

    return True


def print_dependency_graph(
    analysis_id: typing.Optional[str] = None, *, fold: bool = True
) -> bool:
    """Print dependency graph to stdout produced by the given adviser."""
    analysis_id = analysis_id or get_last_analysis_id()
    if not analysis_id.startswith("adviser-"):
        raise UnknownAnalysisType(
            f"An adviser identifier is required to print a dependency graph, got {analysis_id!r} instead"
        )

    adviser_document, error = get_analysis_results(analysis_id)

    if error:
        print("Cannot print dependency graph as advise was not successful")
        return False

    return print_dependency_graph_from_adviser_document(adviser_document, fold=fold)


def add_requirements_to_project(
    requirement: typing.List[str],
    runtime_environment: typing.Optional[str],
    index_url: str,
    dev: bool,
):
    """Add requirements to project."""
    project = thoth_config.get_project(runtime_environment, missing_dir_ok=True)

    for req in requirement:
        _LOGGER.info(
            "Adding %r to %s requirements of runtime environment %r",
            req,
            "development" if dev else "default",
            project.runtime_environment.name,
        )
        project.pipfile.add_requirement(
            req, is_dev=dev, index_url=index_url, force=True
        )
        runtime_environment_config = thoth_config.get_runtime_environment(
            runtime_environment
        )
        python_version = runtime_environment_config.get("python_version")
        if python_version:
            project.set_python_version(python_version)

    _LOGGER.warning(
        "Changes done might require triggering new advise to resolve dependencies"
    )
    thoth_config.save_project(project)


def check_runtime_environment_run(runtime_environment: Dict[str, Any]) -> None:
    """Check the runtime environment matches runtime environment declared in the config file."""
    discovered = discover_all()

    if runtime_environment.get("base_image") != discovered.get("base_image"):
        raise PedanticRunVerificationError(
            f"Base image configured {runtime_environment.get('base_image')!r} does not match "
            f"base image discovered {discovered.get('base_image')!r}"
        )

    if runtime_environment.get("cuda_version") != discovered.get("cuda_version"):
        raise PedanticRunVerificationError(
            f"CUDA version configured {runtime_environment.get('cuda_version')!r} does not match "
            f"CUDA version discovered {discovered.get('cuda_version')!r}"
        )

    if runtime_environment.get("platform", "linux-x86_64") != discovered.get(
        "platform"
    ):
        raise PedanticRunVerificationError(
            f"Platform configured {runtime_environment.get('platform')!r} does not match "
            f"platform discovered {discovered.get('platform')!r}"
        )

    if runtime_environment.get("python_version") != discovered.get("python_version"):
        raise PedanticRunVerificationError(
            f"Python version configured {runtime_environment.get('python_version')!r} does not match "
            f"Python version discovered {discovered.get('python_version')!r}"
        )

    hw_discovered = discovered.get("hardware") or {}
    hw_configured = runtime_environment.get("hardware") or {}
    if hw_configured.get("cpu_family") is not None and hw_configured.get(
        "cpu_family"
    ) != hw_discovered.get("cpu_family"):
        raise PedanticRunVerificationError(
            f"CPU family configured {hw_configured.get('cpu_family')!r} does not match "
            f"CPU family discovered {hw_discovered.get('cpu_family')!r}"
        )

    if hw_configured.get("cpu_model") is not None and hw_configured.get(
        "cpu_model"
    ) != hw_discovered.get("cpu_model"):
        raise PedanticRunVerificationError(
            f"CPU model configured {hw_configured.get('cpu_model')!r} does not match "
            f"CPU model discovered {hw_discovered.get('cpu_model')!r}"
        )

    if hw_configured.get("gpu_model") is not None and hw_configured.get(
        "gpu_model"
    ) != hw_discovered.get("gpu_model"):
        raise PedanticRunVerificationError(
            f"GPU model configured {hw_configured.get('gpu_model')!r} does not match "
            f"GPU model discovered {hw_discovered.get('gpu_model')!r}"
        )

    os_discovered = discovered.get("operating_system") or {}
    os_configured = runtime_environment.get("operating_system") or {}
    if os_configured.get("name") != os_discovered.get("name"):
        raise PedanticRunVerificationError(
            f"Operating system name configured {os_configured.get('name')!r} does not match "
            f"operating system name discovered {os_discovered.get('name')!r}"
        )

    if os_configured.get("version") != os_discovered.get("version"):
        raise PedanticRunVerificationError(
            f"Operating system version configured {os_configured.get('version')!r} does not match "
            f"operating system version discovered {os_discovered.get('version')!r}"
        )

    _LOGGER.info("Runtime environment configured matches runtime environment used")


def load_dot_env(dot_env_path: str) -> Dict[str, str]:
    """Load .env file and parse its content."""
    if not os.path.isfile(dot_env_path):
        _LOGGER.debug("No .env file at %r found", dot_env_path)
        return {}

    _LOGGER.info("Loading environment variables from %r", dot_env_path)

    env = {}
    with open(dot_env_path) as f:
        for line_no, line in enumerate(f.readlines()):
            line = line.strip()
            if not line or line.startswith("#"):  # Skip comments.
                continue

            parts = line.split("=", maxsplit=1)
            if len(parts) != 2:
                _LOGGER.error(
                    "Failed to parse %r content at line %d, ignoring...",
                    dot_env_path,
                    line_no,
                )
                continue

            env[parts[0]] = parts[1]

    return env



# coding: utf-8

"""
    Thoth User API

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: 0.7.0-dev

    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""

import pprint
import re  # noqa: F401

import six

class PythonPackageIndexesIndexes(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """
    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'only_if_package_seen': 'bool',
        'url': 'str',
        'verify_ssl': 'bool',
        'warehouse_api_url': 'str'
    }

    attribute_map = {
        'only_if_package_seen': 'only_if_package_seen',
        'url': 'url',
        'verify_ssl': 'verify_ssl',
        'warehouse_api_url': 'warehouse_api_url'
    }

    def __init__(self, only_if_package_seen=None, url=None, verify_ssl=None, warehouse_api_url=None):  # noqa: E501
        """PythonPackageIndexesIndexes - a model defined in Swagger"""  # noqa: E501
        self._only_if_package_seen = None
        self._url = None
        self._verify_ssl = None
        self._warehouse_api_url = None
        self.discriminator = None
        self.only_if_package_seen = only_if_package_seen
        self.url = url
        self.verify_ssl = verify_ssl
        self.warehouse_api_url = warehouse_api_url

    @property
    def only_if_package_seen(self):
        """Gets the only_if_package_seen of this PythonPackageIndexesIndexes.  # noqa: E501

        The backend is configured to monitor only releases of already seen packages   # noqa: E501

        :return: The only_if_package_seen of this PythonPackageIndexesIndexes.  # noqa: E501
        :rtype: bool
        """
        return self._only_if_package_seen

    @only_if_package_seen.setter
    def only_if_package_seen(self, only_if_package_seen):
        """Sets the only_if_package_seen of this PythonPackageIndexesIndexes.

        The backend is configured to monitor only releases of already seen packages   # noqa: E501

        :param only_if_package_seen: The only_if_package_seen of this PythonPackageIndexesIndexes.  # noqa: E501
        :type: bool
        """
        if only_if_package_seen is None:
            raise ValueError("Invalid value for `only_if_package_seen`, must not be `None`")  # noqa: E501

        self._only_if_package_seen = only_if_package_seen

    @property
    def url(self):
        """Gets the url of this PythonPackageIndexesIndexes.  # noqa: E501

        URL to the Python simple repository as described in PEP 503  # noqa: E501

        :return: The url of this PythonPackageIndexesIndexes.  # noqa: E501
        :rtype: str
        """
        return self._url

    @url.setter
    def url(self, url):
        """Sets the url of this PythonPackageIndexesIndexes.

        URL to the Python simple repository as described in PEP 503  # noqa: E501

        :param url: The url of this PythonPackageIndexesIndexes.  # noqa: E501
        :type: str
        """
        if url is None:
            raise ValueError("Invalid value for `url`, must not be `None`")  # noqa: E501

        self._url = url

    @property
    def verify_ssl(self):
        """Gets the verify_ssl of this PythonPackageIndexesIndexes.  # noqa: E501

        Use secured connection to warehouse  # noqa: E501

        :return: The verify_ssl of this PythonPackageIndexesIndexes.  # noqa: E501
        :rtype: bool
        """
        return self._verify_ssl

    @verify_ssl.setter
    def verify_ssl(self, verify_ssl):
        """Sets the verify_ssl of this PythonPackageIndexesIndexes.

        Use secured connection to warehouse  # noqa: E501

        :param verify_ssl: The verify_ssl of this PythonPackageIndexesIndexes.  # noqa: E501
        :type: bool
        """
        if verify_ssl is None:
            raise ValueError("Invalid value for `verify_ssl`, must not be `None`")  # noqa: E501

        self._verify_ssl = verify_ssl

    @property
    def warehouse_api_url(self):
        """Gets the warehouse_api_url of this PythonPackageIndexesIndexes.  # noqa: E501

        URL to the warehouse API  # noqa: E501

        :return: The warehouse_api_url of this PythonPackageIndexesIndexes.  # noqa: E501
        :rtype: str
        """
        return self._warehouse_api_url

    @warehouse_api_url.setter
    def warehouse_api_url(self, warehouse_api_url):
        """Sets the warehouse_api_url of this PythonPackageIndexesIndexes.

        URL to the warehouse API  # noqa: E501

        :param warehouse_api_url: The warehouse_api_url of this PythonPackageIndexesIndexes.  # noqa: E501
        :type: str
        """

        self._warehouse_api_url = warehouse_api_url

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(PythonPackageIndexesIndexes, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, PythonPackageIndexesIndexes):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        return not self == other



#!/usr/bin/env python3
# thamos
# Copyright(C) 2018 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Utility and helper functions for Thamos."""

from typing import Optional
from contextlib import contextmanager
import logging
import os
import sys

from thoth.common import cwd

from .exceptions import NoProjectDirError

# Limit traversing to parent directories so we handle root - we do not loop over and over in root and we also
# handle cyclic symlinks natively.
_WORKDIR_DEPTH_LEN = 33
_LOGGER = logging.getLogger(__name__)


@contextmanager  # type: ignore
def workdir(file_lookup: Optional[str] = None, warn_on_dir_change: bool = True) -> None:
    """Find project directory and cd into it."""
    file_lookup = file_lookup or ".thoth.yaml"

    project_dir = os.getcwd()
    original_project_dir = project_dir
    for _ in range(_WORKDIR_DEPTH_LEN):
        file = os.path.join(project_dir, file_lookup)
        if os.path.isfile(file):
            with cwd(project_dir):
                if project_dir != original_project_dir and warn_on_dir_change:
                    _LOGGER.warning("Using %r as project root directory", project_dir)
                yield project_dir
            break

        project_dir = os.path.dirname(project_dir)
    else:
        raise NoProjectDirError(
            f"No {file_lookup} found in the current directory {os.getcwd()!r} or in any of its parent "
            f"directories, you can generate it using '{sys.argv[0]} config'"
        )



# coding: utf-8

# flake8: noqa
"""
    Thoth User API

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: 0.7.0-dev

    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""

from __future__ import absolute_import

# import models into model package
from thamos.swagger_client.models.advise_input import AdviseInput
from thamos.swagger_client.models.advise_input_library_usage import AdviseInputLibraryUsage
from thamos.swagger_client.models.adviser_result_response import AdviserResultResponse
from thamos.swagger_client.models.adviser_result_response_metadata import AdviserResultResponseMetadata
from thamos.swagger_client.models.adviser_result_response_metadata_distribution import AdviserResultResponseMetadataDistribution
from thamos.swagger_client.models.adviser_result_response_metadata_distribution_version_parts import AdviserResultResponseMetadataDistributionVersionParts
from thamos.swagger_client.models.adviser_result_response_metadata_python import AdviserResultResponseMetadataPython
from thamos.swagger_client.models.adviser_result_response_result import AdviserResultResponseResult
from thamos.swagger_client.models.adviser_result_response_result_report import AdviserResultResponseResultReport
from thamos.swagger_client.models.adviser_result_response_result_report_dependency_graph import AdviserResultResponseResultReportDependencyGraph
from thamos.swagger_client.models.adviser_result_response_result_report_pipeline import AdviserResultResponseResultReportPipeline
from thamos.swagger_client.models.adviser_result_response_result_report_products import AdviserResultResponseResultReportProducts
from thamos.swagger_client.models.analysis_log_response import AnalysisLogResponse
from thamos.swagger_client.models.analysis_response import AnalysisResponse
from thamos.swagger_client.models.analysis_response_error import AnalysisResponseError
from thamos.swagger_client.models.analysis_result_response import AnalysisResultResponse
from thamos.swagger_client.models.analysis_result_response_metadata import AnalysisResultResponseMetadata
from thamos.swagger_client.models.analysis_result_response_result import AnalysisResultResponseResult
from thamos.swagger_client.models.analysis_result_response_result_aicoeci import AnalysisResultResponseResultAicoeci
from thamos.swagger_client.models.analysis_result_response_result_deb import AnalysisResultResponseResultDeb
from thamos.swagger_client.models.analysis_result_response_result_debdependencies import AnalysisResultResponseResultDebdependencies
from thamos.swagger_client.models.analysis_result_response_result_pythonfiles import AnalysisResultResponseResultPythonfiles
from thamos.swagger_client.models.analysis_result_response_result_pythoninterpreters import AnalysisResultResponseResultPythoninterpreters
from thamos.swagger_client.models.analysis_result_response_result_pythonpackages import AnalysisResultResponseResultPythonpackages
from thamos.swagger_client.models.analysis_status_response import AnalysisStatusResponse
from thamos.swagger_client.models.analysis_status_response_status import AnalysisStatusResponseStatus
from thamos.swagger_client.models.analysis_unauthorized_error import AnalysisUnauthorizedError
from thamos.swagger_client.models.analysis_with_authentication_response import AnalysisWithAuthenticationResponse
from thamos.swagger_client.models.build import Build
from thamos.swagger_client.models.build_analysis_response import BuildAnalysisResponse
from thamos.swagger_client.models.build_analysis_response_base_image_analysis import BuildAnalysisResponseBaseImageAnalysis
from thamos.swagger_client.models.build_analysis_response_buildlog_analysis import BuildAnalysisResponseBuildlogAnalysis
from thamos.swagger_client.models.build_analysis_response_error import BuildAnalysisResponseError
from thamos.swagger_client.models.build_analysis_response_error_base_image_analysis import BuildAnalysisResponseErrorBaseImageAnalysis
from thamos.swagger_client.models.build_analysis_response_error_build_log_analysis import BuildAnalysisResponseErrorBuildLogAnalysis
from thamos.swagger_client.models.build_build_log import BuildBuildLog
from thamos.swagger_client.models.container_images_response import ContainerImagesResponse
from thamos.swagger_client.models.container_images_response_container_images import ContainerImagesResponseContainerImages
from thamos.swagger_client.models.container_images_response_error import ContainerImagesResponseError
from thamos.swagger_client.models.container_images_response_parameters import ContainerImagesResponseParameters
from thamos.swagger_client.models.deb_dependency_item import DebDependencyItem
from thamos.swagger_client.models.image_metadata_response import ImageMetadataResponse
from thamos.swagger_client.models.justification import Justification
from thamos.swagger_client.models.justification_inner import JustificationInner
from thamos.swagger_client.models.kebechet_metadata import KebechetMetadata
from thamos.swagger_client.models.kebechet_webhook_input import KebechetWebhookInput
from thamos.swagger_client.models.project_def import ProjectDef
from thamos.swagger_client.models.project_def_constraints import ProjectDefConstraints
from thamos.swagger_client.models.provenance_input import ProvenanceInput
from thamos.swagger_client.models.provenance_result_response import ProvenanceResultResponse
from thamos.swagger_client.models.provenance_result_response_metadata import ProvenanceResultResponseMetadata
from thamos.swagger_client.models.provenance_result_response_metadata_distribution import ProvenanceResultResponseMetadataDistribution
from thamos.swagger_client.models.provenance_result_response_metadata_distribution_version_parts import ProvenanceResultResponseMetadataDistributionVersionParts
from thamos.swagger_client.models.provenance_result_response_result import ProvenanceResultResponseResult
from thamos.swagger_client.models.provenance_result_response_result_parameters import ProvenanceResultResponseResultParameters
from thamos.swagger_client.models.python_environments import PythonEnvironments
from thamos.swagger_client.models.python_environments_environment import PythonEnvironmentsEnvironment
from thamos.swagger_client.models.python_package_dependencies import PythonPackageDependencies
from thamos.swagger_client.models.python_package_dependencies_dependencies import PythonPackageDependenciesDependencies
from thamos.swagger_client.models.python_package_dependencies_error import PythonPackageDependenciesError
from thamos.swagger_client.models.python_package_indexes import PythonPackageIndexes
from thamos.swagger_client.models.python_package_indexes_indexes import PythonPackageIndexesIndexes
from thamos.swagger_client.models.python_package_name_import_response import PythonPackageNameImportResponse
from thamos.swagger_client.models.python_package_name_import_response_package_names import PythonPackageNameImportResponsePackageNames
from thamos.swagger_client.models.python_package_name_response_error import PythonPackageNameResponseError
from thamos.swagger_client.models.python_package_version_environments_response import PythonPackageVersionEnvironmentsResponse
from thamos.swagger_client.models.python_package_version_metadata_response import PythonPackageVersionMetadataResponse
from thamos.swagger_client.models.python_package_version_metadata_response_error import PythonPackageVersionMetadataResponseError
from thamos.swagger_client.models.python_package_version_metadata_response_metadata import PythonPackageVersionMetadataResponseMetadata
from thamos.swagger_client.models.python_package_version_metadata_response_metadata_importlib_metadata import PythonPackageVersionMetadataResponseMetadataImportlibMetadata
from thamos.swagger_client.models.python_package_version_metadata_response_metadata_importlib_metadata_entry_points import PythonPackageVersionMetadataResponseMetadataImportlibMetadataEntryPoints
from thamos.swagger_client.models.python_package_versions_response import PythonPackageVersionsResponse
from thamos.swagger_client.models.python_package_versions_response_error import PythonPackageVersionsResponseError
from thamos.swagger_client.models.python_package_versions_response_versions import PythonPackageVersionsResponseVersions
from thamos.swagger_client.models.python_packages_response import PythonPackagesResponse
from thamos.swagger_client.models.python_packages_response_packages import PythonPackagesResponsePackages
from thamos.swagger_client.models.python_platforms import PythonPlatforms
from thamos.swagger_client.models.python_stack import PythonStack
from thamos.swagger_client.models.repo_init_input import RepoInitInput
from thamos.swagger_client.models.requirements_def import RequirementsDef
from thamos.swagger_client.models.requirements_lock_def import RequirementsLockDef
from thamos.swagger_client.models.runtime_environment import RuntimeEnvironment
from thamos.swagger_client.models.runtime_environment_hardware import RuntimeEnvironmentHardware
from thamos.swagger_client.models.runtime_environment_operating_system import RuntimeEnvironmentOperatingSystem
from thamos.swagger_client.models.stack_info import StackInfo
from thamos.swagger_client.models.unit_def import UnitDef



# coding: utf-8

"""
    Thoth User API

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: 0.7.0-dev

    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""

import pprint
import re  # noqa: F401

import six

class ProvenanceResultResponseResult(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """
    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'error': 'bool',
        'error_msg': 'str',
        'input': 'object',
        'parameters': 'ProvenanceResultResponseResultParameters',
        'report': 'list[object]'
    }

    attribute_map = {
        'error': 'error',
        'error_msg': 'error_msg',
        'input': 'input',
        'parameters': 'parameters',
        'report': 'report'
    }

    def __init__(self, error=None, error_msg=None, input=None, parameters=None, report=None):  # noqa: E501
        """ProvenanceResultResponseResult - a model defined in Swagger"""  # noqa: E501
        self._error = None
        self._error_msg = None
        self._input = None
        self._parameters = None
        self._report = None
        self.discriminator = None
        self.error = error
        self.error_msg = error_msg
        self.input = input
        self.parameters = parameters
        self.report = report

    @property
    def error(self):
        """Gets the error of this ProvenanceResultResponseResult.  # noqa: E501

        A flag set to true if an error was encountered  # noqa: E501

        :return: The error of this ProvenanceResultResponseResult.  # noqa: E501
        :rtype: bool
        """
        return self._error

    @error.setter
    def error(self, error):
        """Sets the error of this ProvenanceResultResponseResult.

        A flag set to true if an error was encountered  # noqa: E501

        :param error: The error of this ProvenanceResultResponseResult.  # noqa: E501
        :type: bool
        """
        if error is None:
            raise ValueError("Invalid value for `error`, must not be `None`")  # noqa: E501

        self._error = error

    @property
    def error_msg(self):
        """Gets the error_msg of this ProvenanceResultResponseResult.  # noqa: E501

        An error message reported to users  # noqa: E501

        :return: The error_msg of this ProvenanceResultResponseResult.  # noqa: E501
        :rtype: str
        """
        return self._error_msg

    @error_msg.setter
    def error_msg(self, error_msg):
        """Sets the error_msg of this ProvenanceResultResponseResult.

        An error message reported to users  # noqa: E501

        :param error_msg: The error_msg of this ProvenanceResultResponseResult.  # noqa: E501
        :type: str
        """

        self._error_msg = error_msg

    @property
    def input(self):
        """Gets the input of this ProvenanceResultResponseResult.  # noqa: E501


        :return: The input of this ProvenanceResultResponseResult.  # noqa: E501
        :rtype: object
        """
        return self._input

    @input.setter
    def input(self, input):
        """Sets the input of this ProvenanceResultResponseResult.


        :param input: The input of this ProvenanceResultResponseResult.  # noqa: E501
        :type: object
        """

        self._input = input

    @property
    def parameters(self):
        """Gets the parameters of this ProvenanceResultResponseResult.  # noqa: E501


        :return: The parameters of this ProvenanceResultResponseResult.  # noqa: E501
        :rtype: ProvenanceResultResponseResultParameters
        """
        return self._parameters

    @parameters.setter
    def parameters(self, parameters):
        """Sets the parameters of this ProvenanceResultResponseResult.


        :param parameters: The parameters of this ProvenanceResultResponseResult.  # noqa: E501
        :type: ProvenanceResultResponseResultParameters
        """
        if parameters is None:
            raise ValueError("Invalid value for `parameters`, must not be `None`")  # noqa: E501

        self._parameters = parameters

    @property
    def report(self):
        """Gets the report of this ProvenanceResultResponseResult.  # noqa: E501

        A provenance check report  # noqa: E501

        :return: The report of this ProvenanceResultResponseResult.  # noqa: E501
        :rtype: list[object]
        """
        return self._report

    @report.setter
    def report(self, report):
        """Sets the report of this ProvenanceResultResponseResult.

        A provenance check report  # noqa: E501

        :param report: The report of this ProvenanceResultResponseResult.  # noqa: E501
        :type: list[object]
        """
        if report is None:
            raise ValueError("Invalid value for `report`, must not be `None`")  # noqa: E501

        self._report = report

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(ProvenanceResultResponseResult, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ProvenanceResultResponseResult):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        return not self == other



# coding: utf-8

"""
    Thoth User API

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: 0.7.0-dev

    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""

import pprint
import re  # noqa: F401

import six

class AdviserResultResponseResultReportPipeline(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """
    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'boots': 'list[UnitDef]',
        'pseudonyms': 'list[UnitDef]',
        'sieves': 'list[UnitDef]',
        'steps': 'list[UnitDef]',
        'strides': 'list[UnitDef]',
        'wraps': 'list[UnitDef]'
    }

    attribute_map = {
        'boots': 'boots',
        'pseudonyms': 'pseudonyms',
        'sieves': 'sieves',
        'steps': 'steps',
        'strides': 'strides',
        'wraps': 'wraps'
    }

    def __init__(self, boots=None, pseudonyms=None, sieves=None, steps=None, strides=None, wraps=None):  # noqa: E501
        """AdviserResultResponseResultReportPipeline - a model defined in Swagger"""  # noqa: E501
        self._boots = None
        self._pseudonyms = None
        self._sieves = None
        self._steps = None
        self._strides = None
        self._wraps = None
        self.discriminator = None
        self.boots = boots
        self.pseudonyms = pseudonyms
        self.sieves = sieves
        self.steps = steps
        self.strides = strides
        self.wraps = wraps

    @property
    def boots(self):
        """Gets the boots of this AdviserResultResponseResultReportPipeline.  # noqa: E501

        Boot pipeline unit types registered to the resolution process   # noqa: E501

        :return: The boots of this AdviserResultResponseResultReportPipeline.  # noqa: E501
        :rtype: list[UnitDef]
        """
        return self._boots

    @boots.setter
    def boots(self, boots):
        """Sets the boots of this AdviserResultResponseResultReportPipeline.

        Boot pipeline unit types registered to the resolution process   # noqa: E501

        :param boots: The boots of this AdviserResultResponseResultReportPipeline.  # noqa: E501
        :type: list[UnitDef]
        """
        if boots is None:
            raise ValueError("Invalid value for `boots`, must not be `None`")  # noqa: E501

        self._boots = boots

    @property
    def pseudonyms(self):
        """Gets the pseudonyms of this AdviserResultResponseResultReportPipeline.  # noqa: E501

        Pseudonym pipeline unit types registered to the resolution process   # noqa: E501

        :return: The pseudonyms of this AdviserResultResponseResultReportPipeline.  # noqa: E501
        :rtype: list[UnitDef]
        """
        return self._pseudonyms

    @pseudonyms.setter
    def pseudonyms(self, pseudonyms):
        """Sets the pseudonyms of this AdviserResultResponseResultReportPipeline.

        Pseudonym pipeline unit types registered to the resolution process   # noqa: E501

        :param pseudonyms: The pseudonyms of this AdviserResultResponseResultReportPipeline.  # noqa: E501
        :type: list[UnitDef]
        """
        if pseudonyms is None:
            raise ValueError("Invalid value for `pseudonyms`, must not be `None`")  # noqa: E501

        self._pseudonyms = pseudonyms

    @property
    def sieves(self):
        """Gets the sieves of this AdviserResultResponseResultReportPipeline.  # noqa: E501

        Sieve pipeline unit types registered to the resolution process   # noqa: E501

        :return: The sieves of this AdviserResultResponseResultReportPipeline.  # noqa: E501
        :rtype: list[UnitDef]
        """
        return self._sieves

    @sieves.setter
    def sieves(self, sieves):
        """Sets the sieves of this AdviserResultResponseResultReportPipeline.

        Sieve pipeline unit types registered to the resolution process   # noqa: E501

        :param sieves: The sieves of this AdviserResultResponseResultReportPipeline.  # noqa: E501
        :type: list[UnitDef]
        """
        if sieves is None:
            raise ValueError("Invalid value for `sieves`, must not be `None`")  # noqa: E501

        self._sieves = sieves

    @property
    def steps(self):
        """Gets the steps of this AdviserResultResponseResultReportPipeline.  # noqa: E501

        Step pipeline unit types registered to the resolution process   # noqa: E501

        :return: The steps of this AdviserResultResponseResultReportPipeline.  # noqa: E501
        :rtype: list[UnitDef]
        """
        return self._steps

    @steps.setter
    def steps(self, steps):
        """Sets the steps of this AdviserResultResponseResultReportPipeline.

        Step pipeline unit types registered to the resolution process   # noqa: E501

        :param steps: The steps of this AdviserResultResponseResultReportPipeline.  # noqa: E501
        :type: list[UnitDef]
        """
        if steps is None:
            raise ValueError("Invalid value for `steps`, must not be `None`")  # noqa: E501

        self._steps = steps

    @property
    def strides(self):
        """Gets the strides of this AdviserResultResponseResultReportPipeline.  # noqa: E501

        Stride pipeline unit types registered to the resolution process   # noqa: E501

        :return: The strides of this AdviserResultResponseResultReportPipeline.  # noqa: E501
        :rtype: list[UnitDef]
        """
        return self._strides

    @strides.setter
    def strides(self, strides):
        """Sets the strides of this AdviserResultResponseResultReportPipeline.

        Stride pipeline unit types registered to the resolution process   # noqa: E501

        :param strides: The strides of this AdviserResultResponseResultReportPipeline.  # noqa: E501
        :type: list[UnitDef]
        """
        if strides is None:
            raise ValueError("Invalid value for `strides`, must not be `None`")  # noqa: E501

        self._strides = strides

    @property
    def wraps(self):
        """Gets the wraps of this AdviserResultResponseResultReportPipeline.  # noqa: E501

        Wrap pipeline unit types registered to the resolution process   # noqa: E501

        :return: The wraps of this AdviserResultResponseResultReportPipeline.  # noqa: E501
        :rtype: list[UnitDef]
        """
        return self._wraps

    @wraps.setter
    def wraps(self, wraps):
        """Sets the wraps of this AdviserResultResponseResultReportPipeline.

        Wrap pipeline unit types registered to the resolution process   # noqa: E501

        :param wraps: The wraps of this AdviserResultResponseResultReportPipeline.  # noqa: E501
        :type: list[UnitDef]
        """
        if wraps is None:
            raise ValueError("Invalid value for `wraps`, must not be `None`")  # noqa: E501

        self._wraps = wraps

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(AdviserResultResponseResultReportPipeline, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, AdviserResultResponseResultReportPipeline):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        return not self == other
